SystemOrganization addCategory: #'TrueType-Fonts'!
SystemOrganization addCategory: #'TrueType-Support'!

Object subclass: #TTContourConstruction
	instanceVariableNames: 'points'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTContourConstruction commentStamp: '<historical>' prior: 0!
This class represents a temporary contour structure during the construction of a TTGlyph from a TrueType file.

Instance variables:
	points	<Array of: TTPoint>	The points defining this contour!

!TTContourConstruction class methodsFor: 'instance creation' stamp: ''!
on: points

	^self new points: points! !

!TTContourConstruction methodsFor: 'converting' stamp: 'di 11/21/1999 20:19'!
asCompressedPoints
	"Return the receiver compressed into a PointArray.
	All lines will be converted into bezier segments with
	the control point set to the start point"
	| out minPt maxPt fullRange |
	minPt := -16r7FFF asPoint.
	maxPt := 16r8000 asPoint.
	"Check if we need full 32bit range"
	fullRange := points anySatisfy: [:any| any asPoint < minPt or:[any asPoint > maxPt]].
	fullRange ifTrue:[
		out := WriteStream on: (PointArray new: points size).
	] ifFalse:[
		out := WriteStream on: (ShortPointArray new: points size).
	].
	self segmentsDo:[:segment|
		out nextPut: segment start.
		segment isBezier2Segment 
			ifTrue:[out nextPut: segment via]
			ifFalse:[out nextPut: segment start].
		out nextPut: segment end.
	].
	^out contents! !

!TTContourConstruction methodsFor: 'accessing' stamp: ''!
points
	^points! !

!TTContourConstruction methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:34'!
points: anArray
	points := anArray asArray.! !

!TTContourConstruction methodsFor: 'printing' stamp: ''!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: points size;
		"space;
		print: self type;"
		nextPut:$)! !

!TTContourConstruction methodsFor: 'accessing' stamp: ''!
segments

	| segments |
	segments := OrderedCollection new.
	self segmentsDo:[:seg| segments add: seg].
	^segments! !

!TTContourConstruction methodsFor: 'enumerating' stamp: 'hmm 10/28/2001 21:55'!
segmentsDo: aBlock
	"Evaluate aBlock with the segments of the receiver. This may either be straight line
	segments or quadratic bezier curves. The decision is made upon the type flags
	in TTPoint as follows:
	a) 	Two subsequent #OnCurve points define a straight segment
	b) 	An #OnCurve point followed by an #OffCurve point followed 
		by an #OnCurve point defines a quadratic bezier segment
	c)	Two subsequent #OffCurve points have an implicitely defined 
		#OnCurve point at half the distance between them"
	| last next mid index i |
	last := points first.
	"Handle case where first point is off-curve"
	(last type == #OnCurve) ifFalse: [
		i := points findFirst: [:pt | pt type == #OnCurve].
		i = 0
			ifTrue: [mid := TTPoint new
							type: #OnCurve;
							x: points first x + points last x // 2;
							y: points first y + points last y // 2.
					points := (Array with: mid), points]
			ifFalse: [points := (points copyFrom: i to: points size), (points copyFrom: 1 to: i)].
		last := points first].
	index := 2.
	[index <= points size] whileTrue:[
		mid := points at: index.
		mid type == #OnCurve ifTrue:[
			"Straight segment"
			aBlock value: (LineSegment from: last asPoint to: mid asPoint).
			last := mid.
		] ifFalse:["Quadratic bezier"
			"Read ahead if the next point is on curve"
			next := (index < points size) ifTrue:[points at: (index+1)] ifFalse:[points first].
			next type == #OnCurve ifTrue:[
				"We'll continue after the end point"
				index := index + 1.
			] ifFalse:[ "Calculate center"
				next := (next asPoint + mid asPoint) // 2].
			aBlock value:(Bezier2Segment from: last asPoint via: mid asPoint to: next asPoint).
			last := next].
		index := index + 1].
	(index = (points size + 1)) ifTrue:[
		aBlock value:(LineSegment from: points last asPoint to: points first asPoint)]! !

Object subclass: #TTFileDescription
	instanceVariableNames: 'fileName fileOffset familyName subfamilyName copyright ascender descender lineGap unitsPerEm numGlyphs indexToLocOffset indexToLocFormat glyphTableOffset cmapType cmapOffset numHMetrics hmtxTableOffset sTypoAscender sTypoDescender sTypoLineGap'
	classVariableNames: 'AllFontsAndFiles FontPaths OfferNonPortableFonts'
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTFileDescription commentStamp: 'ar 7/29/2009 22:18' prior: 0!
Contrary to TTFontDescritption, this class leaves true type files on disk and only reads the required portions when constructing glyphs. This avoids the need of reading the entire font into memory at the cost of having to hit disk whenever a glyph is requested.!

!TTFileDescription class methodsFor: 'font paths' stamp: 'nice 1/19/2010 02:04'!
allFamilyNamesAndFiles
	"Answer a dictionary of all known family names and their corresponding file names."
	
	AllFontsAndFiles ifNil:[
		AllFontsAndFiles := Dictionary new.
		Cursor wait showWhile:[self allFontsDo:[:font|
			| names |
			names := AllFontsAndFiles at: font familyName 
					ifAbsentPut:[OrderedCollection new].
			names add: font fileName]]].
	^AllFontsAndFiles ! !

!TTFileDescription class methodsFor: 'font paths' stamp: 'nice 1/19/2010 02:03'!
allFontsAndFiles
	"Answer a dictionary of all known family names and their corresponding file names."
	
	AllFontsAndFiles ifNil:[
		AllFontsAndFiles := Dictionary new.
		Cursor wait showWhile:[self allFontsDo:[:font|
			| names |
			names := AllFontsAndFiles at: font familyName 
					ifAbsentPut:[OrderedCollection new].
			names add: font fileName]]].
	^AllFontsAndFiles ! !

!TTFileDescription class methodsFor: 'font paths' stamp: 'ar 7/30/2009 20:37'!
allFontsDo: aBlock
	"Evaluate aBlock with all the fonts we can find. Use sparingly."
	self fontPathsDo:[:path|
		self fontFilesIn: path do:[:font|
			font familyName 
				ifNotNil:[aBlock value: font]]]! !

!TTFileDescription class methodsFor: 'font paths' stamp: 'nice 1/18/2010 19:17'!
findFontFile: fontFileName
	"Find the path containing the font with the given name.
	If it can't be found, return nil."
	
	self fontPathsDo:[:path|
		| fd |
		fd := FileDirectory on: path.
		([fd fileExists: fontFileName] on: Error do:[false]) 
			ifTrue:[^fd fullNameFor: fontFileName].
	].
	^nil! !

!TTFileDescription class methodsFor: 'font paths' stamp: 'ul 8/2/2011 20:11'!
fontFilesIn: path do: aBlock
	"TTFileDescription loadAllFilesIn: 'C:\Windows\Fonts'"
	"Load all the TTF files we can find in the given path"
	| fd |
	fd := FileDirectory on: path.
	(fd fileNamesMatching: '*.ttf;*.ttc') do:[:fn|
		(self readFontsFrom: fn) do: aBlock ]! !

!TTFileDescription class methodsFor: 'user interaction' stamp: 'ar 7/30/2009 20:22'!
fontFromUser
	"TTFileDescription fontFromUser"
	^self fontFromUser: TextStyle defaultFont! !

!TTFileDescription class methodsFor: 'user interaction' stamp: 'ar 7/30/2009 20:16'!
fontFromUser: priorFont
	^self fontFromUser: priorFont allowKeyboard: true! !

!TTFileDescription class methodsFor: 'user interaction' stamp: 'nice 12/27/2009 03:12'!
fontFromUser: priorFont allowKeyboard: aBoolean
	"TTFileDescription fontFromUser"
	| fontMenu fontNames builder resultBlock style font widget result |
	builder := ToolBuilder default.
	fontNames := self allFontsAndFiles keys asArray sort.
	fontMenu := builder pluggableMenuSpec new.
	fontMenu label: 'Non-portable fonts'.
	resultBlock := [:value| result := value].
	fontNames do: [:fontName | | active ptMenu item |
		active := priorFont familyName sameAs: fontName.
		ptMenu := builder pluggableMenuSpec new.
		TTCFont pointSizes do: [:pt | | label |
			label := pt printString, ' pt'.
			item := ptMenu add: label 
				target: resultBlock
				selector: #value:
				argumentList: {{fontName. pt}}.
			item checked: (active and:[pt = priorFont pointSize]).
		].
		item := fontMenu add: fontName action: nil.
		item subMenu: ptMenu.
		item checked: active.
	].
	widget := builder open: fontMenu.
	builder runModal: widget.
	result ifNil:[^nil].
	style := (TextStyle named: result first) ifNil:[self installFamilyNamed: result first].
	style ifNil: [^ self].
	font := style fonts detect: [:any | any pointSize = result last] ifNone: [nil].
	^ font
! !

!TTFileDescription class methodsFor: 'instance creation' stamp: 'nice 3/28/2010 19:00'!
fontOffsetsInFile: file
	"Answer a collection of font offsets in the given file"
	| tag version nFonts |
	file position: 0.
	tag := file next: 4.
	tag caseOf: {
		['true' asByteArray] -> ["Version 1.0 TTF file"
			"http://developer.apple.com/textfonts/TTRefMan/RM06/Chap6.html
			The values 'true' (0x74727565) and 0x00010000 are recognized by the Mac OS 
			as referring to TrueType fonts."
			^Array with: 0 "only one font"
		].
		[#[0 1 0 0]] -> ["Version 1.0 TTF file"
			^Array with: 0 "only one font"
		].
		['ttcf' asByteArray]	-> ["TTC file"
			version := file next: 4.
			(version = #[0 1 0 0] or: [version = #[0 2 0 0]]) ifFalse: [^self error: 'Unsupported TTC version'].
			nFonts := file nextNumber: 4.
			^(1 to: nFonts) collect: [:i | file nextNumber: 4].
		].
	} otherwise:[
		self error: 'This is not a valid Truetype file'.
	].! !

!TTFileDescription class methodsFor: 'font paths' stamp: 'topa 3/6/2015 22:07'!
fontPathsDo: aBlock
	"Evaluate aBlock with all of the font paths that should be searched on the current platform"

	"Start with the current directory"
	aBlock value: FileDirectory default pathName.

	"Then subdirectory 'fonts'"
	aBlock value: (FileDirectory default directoryNamed: 'fonts') pathName.

	"Platform specific directories"
	Smalltalk platformName caseOf:{
		['Win32']	->	[
			"Standard Windows fonts directory"
			aBlock value: 'C:\Windows\Fonts'.
		].
		['Mac OS']	->	[
			"Standard system fonts directories"
			#('/System/Library/Fonts' '/Library/Fonts') do: [:fontDir |
				aBlock value: fontDir].
		].
		['unix']		->	[ | base |
			"Standard fonts are in /usr/share/fonts/*"
			base := '/usr/share/fonts'.
			(FileDirectory on: base) directoryNames 
				do:[:dn| aBlock value: base, '/', dn].
		].
	} otherwise:[].
	! !

!TTFileDescription class methodsFor: 'class initialization' stamp: 'ar 7/30/2009 21:31'!
initialize
	"TTFileDescription initialize"
	Smalltalk addToShutDownList: self.
	FontPaths := Dictionary new.
	AllFontsAndFiles := nil.! !

!TTFileDescription class methodsFor: 'instance creation' stamp: 'nice 12/27/2009 03:12'!
installFamilyNamed: familyName
	"Install all the corresponding fonts for this family"
	"
		TTFileDescription installFamilyNamed: 'Arial'.
		TTFileDescription installFamilyNamed: 'Batang'.
	"
	| fontFiles |
	fontFiles := self allFontsAndFiles at: familyName ifAbsent:[#()].
	fontFiles do:[:fileName| | ttDesc |
		ttDesc := (self readFontsFrom: fileName) detect:[:fnt| fnt familyName = familyName].
		TTCFont newTextStyleFromTT: ttDesc.
	].
	^TextStyle named: familyName! !

!TTFileDescription class methodsFor: 'examples' stamp: 'ar 7/30/2009 20:49'!
loadAllFontFiles
	"Load all the TTF files we can find in all font paths"
	"
		TTFileDescription loadAllFontFiles.
	"
	self fontPathsDo:[:path| | fd |
		fd := FileDirectory on: path.
		(fd fileNamesMatching: '*.ttf;*.ttc') do:[:fn|
			(self readFontsFrom: fn) do:[:font|
				(1 to: font numGlyphs) 
					do:[:i| font readGlyphAt: i-1] 
					displayingProgress: 'Reading ', font name].
		] displayingProgress: 'Scanning ', path.
	].! !

!TTFileDescription class methodsFor: 'user interaction' stamp: 'ar 7/30/2009 21:30'!
offerNonPortableFonts
	"Should native fonts be offered when displaying font menus?"
	<preference: 'Offer Native Fonts'
		category: 'Morphic'
		description: 'When true, an additional menu is offered for choosing non-portable fonts'
		type: #Boolean>
	^OfferNonPortableFonts ifNil:[true]! !

!TTFileDescription class methodsFor: 'user interaction' stamp: 'ar 7/30/2009 21:29'!
offerNonPortableFonts: aBool
	"Should native fonts be offered when displaying font menus?"
	OfferNonPortableFonts := aBool.! !

!TTFileDescription class methodsFor: 'instance creation' stamp: 'nice 1/18/2010 19:18'!
openFontFile: fontFileName do: aBlock
	"Open the font with the given font file name"
	| fontFilePath file |
	fontFilePath := FontPaths at: fontFileName 
		ifAbsentPut:[self findFontFile: fontFileName].
	fontFilePath ifNil:[^nil].
	file := [FileStream readOnlyFileNamed: fontFilePath] on: Error do:[:ex|
		"We lost the font; someone might have moved it away"
		fontFilePath removeKey: fontFileName ifAbsent:[].
		^nil
	].
	^[aBlock value: file binary] ensure:[file close].! !

!TTFileDescription class methodsFor: 'instance creation' stamp: 'ar 7/31/2009 22:23'!
readFontsFrom: aFilename
	"Reads and returns all the fonts in the given file"
	"
		TTFileDescription readFontsFrom: 'batang.ttc'.
	"
	^self openFontFile: aFilename do:[:file|
		(self fontOffsetsInFile: file)
			collect:[:offset| self new on: aFilename offset: offset]
			thenSelect:[:font| font notNil]].
	! !

!TTFileDescription class methodsFor: 'class initialization' stamp: 'ar 7/30/2009 21:31'!
shutDown
	"Flush my caches"
	FontPaths := Dictionary new.
	AllFontsAndFiles := nil.! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:07'!
ascender
	"Ascender of the font. Relative to unitsPerEm.
	Easily confused with the typographic ascender."
	^ascender! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/29/2009 23:33'!
at: charOrCode
	"Compatibility with TTFontDescription"
	^self glyphAt: charOrCode! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/30/2009 17:05'!
childGlyphAt: glyphIndex in: glyphCache fromFile: fontFile
	"Get the glyph with the given glyph index. Look in cache first, then read from file.
	Ensure file is positioned at point where it was when it came here."
	^glyphCache at: glyphIndex ifAbsentPut:[ | glyph filePos |
		filePos := fontFile position.
		glyph := self readGlyphAt: glyphIndex fromFile: fontFile.
		fontFile position: filePos.
		glyph].! !

!TTFileDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 22:31'!
copyright

	^ copyright! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:08'!
descender
	"Descender of the font. Relative to unitsPerEm.
	Easily confused with the typographic descender."
	^descender! !

!TTFileDescription methodsFor: 'private' stamp: 'ar 7/30/2009 18:15'!
displayAll
	"Read all the glyphs and display them"
	| glyph form scale points x y |
	points := 24.
	scale := points asFloat / unitsPerEm.
	x := y := 0.
	Display deferUpdates: true.
	1 to: numGlyphs do:[:i|
		glyph := self readGlyphAt: i-1.
		form := glyph asFormWithScale: scale
			ascender: ascender
			descender: descender.
		Display fillWhite: (x@y extent: form extent).
		form displayOn: Display at: x@y rule: 34.
		Display forceToScreen: (x@y extent: form extent).
		x := x + form width.
		x > Display width ifTrue:[y := y + form height. x := 0].
		y > Display height ifTrue:[y := 0].
		Sensor anyButtonPressed ifTrue:[^Display restore].
	].! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'topa 3/9/2015 22:59'!
fallbackGlyph
	"Answer the fallback glyph, the first in the loca table "
	| glyph |
	self withFileDo:[:fontFile| 
		glyph := self readGlyphAt: 0 fromFile: fontFile.
		self updateGlyphMetrics: glyph fromFile: fontFile.
	].
	^glyph! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/29/2009 23:52'!
familyName
	"The family name for the font"
	^familyName! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/30/2009 20:23'!
fileName
	"The name of the Truetype file"
	^fileName! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 7/30/2009 13:55'!
findTable: tag in: fontFile
	"Position the fontFile at the beginning of the table with the given tag.
	Answer true if we found the table, false otherwise."
	| maxTables chksum offset length table |
	fontFile position: fileOffset.
	fontFile skip: 4. "version"
	maxTables := fontFile nextNumber: 2.
	fontFile skip: 6.
	1 to: maxTables do:[:i|
		table := (fontFile next: 4) asString.
		chksum := fontFile nextNumber: 4.
		offset := fontFile nextNumber: 4.
		length := fontFile nextNumber: 4.
		table = tag ifTrue:[
			fontFile position: offset.
			^true].
	].
	chksum. length. "fake usage"
	^false! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 18:46'!
fontHeight
	^ascender - descender! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/29/2009 21:41'!
getGlyphFlagsFrom: fontFile size: nPts
	"Read in the flags for this glyph.  The outer loop gathers the flags that
	are actually contained in the table.  If the repeat bit is set in a flag
	then the next byte is read from the table; this is the number of times
	to repeat the last flag.  The inner loop does this, incrementing the
	outer loops index each time."
	| flags index repCount flagBits |
	flags := ByteArray new: nPts.
	index := 1.
	[index <= nPts] whileTrue:[
		flagBits := fontFile next.
		flags at: index put: flagBits.
		(flagBits bitAnd: 8) = 8 ifTrue:[
			repCount := fontFile next.
			repCount timesRepeat:[
				index := index + 1.
				flags at: index put: flagBits]].
		index := index + 1].
	^flags! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'nice 12/27/2009 03:12'!
glyphAt: charOrCode
	"Answer the glyph with the given code point"
	| codePoint glyph |
	codePoint := charOrCode asCharacter charCode.
	self withFileDo:[:fontFile| | glyphIndex |
		glyphIndex := self readCmapTableAt: codePoint fromFile: fontFile.
		glyph := self readGlyphAt: glyphIndex fromFile: fontFile.
		self updateGlyphMetrics: glyph fromFile: fontFile.
	].
	^glyph! !

!TTFileDescription methodsFor: 'testing' stamp: 'topa 3/19/2015 11:11'!
isRegular
	"Answer true if I am a Regular/Roman font (i.e. not bold, etc.)"
	^ (TTCFont indexOfSubfamilyName: (self subfamilyName)) = 0.
! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:08'!
lineGap
	"Leading of the font. Relative to unitsPerEm.
	Easily confused with the typographic linegap."
	^lineGap! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/30/2009 14:16'!
name
	"For compatibility with TTFontDescription"
	^familyName! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/29/2009 23:32'!
numGlyphs
	"The number of glyphs represented in this font"
	^numGlyphs! !

!TTFileDescription methodsFor: 'initialize' stamp: 'ar 7/29/2009 23:46'!
on: aFileName
	"Initialize the receiver from a file name"
	fileName := aFileName.
	self withFileDo:[:fontFile|
		(self findTable: 'head' in: fontFile) 
			ifFalse:[^self error: 'File does not have a header table'].
		self processFontHeaderTable: fontFile.
		(self findTable: 'maxp' in: fontFile) 
			ifFalse:[^self error: 'File does not have a profile table'].
		self processMaximumProfileTable: fontFile.
		(self findTable: 'name' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		self processNamingTable: fontFile.
		(self findTable: 'hhea' in: fontFile) 
			ifFalse:[^self error: 'File does not have a horizontal header table'].
		self processHorizontalHeaderTable: fontFile.
		(self findTable: 'hmtx' in: fontFile) 
			ifFalse:[^self error: 'File does not have a horizontal header table'].
		hmtxTableOffset := fontFile position.
		(self findTable: 'loca' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		indexToLocOffset := fontFile position.
		(self findTable: 'glyf' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		glyphTableOffset := fontFile position.
		(self findTable: 'cmap' in: fontFile) 
			ifFalse:[^self error: 'File does not have a header table'].
		self processCharacterMappingTable: fontFile.
	].! !

!TTFileDescription methodsFor: 'initialize' stamp: 'topa 3/6/2015 17:06'!
on: aFileName offset: fontOffset
	"Initialize the receiver from a file name"
	fileName := aFileName.
	fileOffset := fontOffset.
	self withFileDo:[:fontFile|
		"Some TTC fonts may actually be collection of PostScript-Based OpenType fonts"
		(self findTable: 'CFF ' in: fontFile)
			ifTrue: [^ nil]
			ifFalse: [fontFile position: fileOffset "reset"].			
		"Some bitmap fonts are called .ttf; skip anything that doesn't have a header"
		(self findTable: 'head' in: fontFile) ifFalse:[^nil].
		self processFontHeaderTable: fontFile.
		(self findTable: 'maxp' in: fontFile) 
			ifFalse:[^self error: 'File does not have a profile table'].
		self processMaximumProfileTable: fontFile.
		(self findTable: 'name' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		self processNamingTable: fontFile.
		(self findTable: 'hhea' in: fontFile) 
			ifFalse:[^self error: 'File does not have a horizontal header table'].
		self processHorizontalHeaderTable: fontFile.
		(self findTable: 'OS/2' in: fontFile)
			ifTrue:[self processOS2Table: fontFile].
		(self findTable: 'hmtx' in: fontFile) 
			ifFalse:[^self error: 'File does not have a horizontal header table'].
		hmtxTableOffset := fontFile position.
		(self findTable: 'loca' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		indexToLocOffset := fontFile position.
		(self findTable: 'glyf' in: fontFile) 
			ifFalse:[^self error: 'File does not have a naming table'].
		glyphTableOffset := fontFile position.
		(self findTable: 'cmap' in: fontFile) 
			ifFalse:[^self error: 'File does not have a header table'].
		self processCharacterMappingTable: fontFile.
	].! !

!TTFileDescription methodsFor: 'printing' stamp: 'ar 7/30/2009 13:57'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: '('; print: fileName; nextPutAll: ')'.! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 7/29/2009 22:29'!
processCharacterMappingTable: fontFile
	"Read the font's character to glyph index mapping table."
	| initialOffset nSubTables pID sID offset |
	initialOffset := fontFile position.
	fontFile skip: 2. "Skip table version"
	nSubTables := fontFile nextNumber: 2.
	1 to: nSubTables do:[:i|
		pID := fontFile nextNumber: 2.
		sID := fontFile nextNumber: 2.
		offset := fontFile nextNumber: 4.
		"Check if this is either a Unicode (0), Macintosh (1),
		or a Windows (3) encoded table"
		(#(0 1 3) includes: pID) ifTrue:[
			cmapType := pID.
			cmapOffset := initialOffset + offset.
			cmapType = 0 ifTrue:[^self]. "found Unicode table; use it"
		].
	].! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/30/2009 17:03'!
processCompositeGlyph: glyph contours: nContours from: fontFile
	"Read a composite glyph from the font data. The glyph passed into this method contains some state variables that must be copied into the resulting composite glyph."
	| flags glyphIndex hasInstr ofsX ofsY iLen a11 a12 a21 a22 m glyphCache |
	glyphCache := Dictionary new.
	a11 := a22 := 16r4000.	"1.0 in F2Dot14"
	a21 := a12 := 0.			"0.0 in F2Dot14"
	"Copy state"
	hasInstr := false.
	[ flags := fontFile nextNumber: 2.
	glyphIndex := fontFile nextNumber: 2.
	(flags bitAnd: 1) = 1 ifTrue:[
		ofsX := self short: (fontFile nextNumber: 2).
		ofsY := self short: (fontFile nextNumber: 2).
	] ifFalse:[
		(ofsX := fontFile next) > 127 ifTrue:[ofsX := ofsX - 256].
		(ofsY := fontFile next) > 127 ifTrue:[ofsY := ofsY - 256].
	].
	((flags bitAnd: 2) = 2) ifFalse:[
		| i1 i2 p1 p2 |
		(flags bitAnd: 1) = 1 ifTrue: [
			i1 := ofsX + 65536 \\ 65536.
			i2 := ofsY + 65536 \\ 65536]
		 ifFalse: [
			i1 := ofsX + 256 \\ 256.
			i2 := ofsY + 256 \\ 256].
		p1 := glyph referenceVertexAt: i1+1.
		p2 := (self childGlyphAt: glyphIndex in: glyphCache fromFile: fontFile) referenceVertexAt: i2+1.
		ofsX := p1 x - p2 x.
		ofsY := p1 y - p2 y.
	].
	(flags bitAnd: 8) = 8 ifTrue:[
		a11 := a22 := self short: (fontFile nextNumber: 2)].
	(flags bitAnd: 64) = 64 ifTrue:[
		a11 := self short: (fontFile nextNumber: 2).
		a22 := self short: (fontFile nextNumber: 2).
	].
	(flags bitAnd: 128) = 128 ifTrue:[
		"2x2 transformation"
		a11 := self short: (fontFile nextNumber: 2).
		a21 := self short: (fontFile nextNumber: 2).
		a12 := self short: (fontFile nextNumber: 2).
		a22 := self short: (fontFile nextNumber: 2).
	].
	m := MatrixTransform2x3 new.
	"Convert entries from F2Dot14 to float"
	m a11: (a11 asFloat / 16r4000).
	m a12: (a12 asFloat / 16r4000).
	m a21: (a21 asFloat / 16r4000).
	m a22: (a22 asFloat / 16r4000).
	m a13: ofsX.
	m a23: ofsY.
	glyph addGlyph: (self childGlyphAt: glyphIndex in: glyphCache fromFile: fontFile) transformation: m.
	hasInstr := hasInstr or:[ (flags bitAnd: 256) = 256].
	"Continue as long as the MORE:=COMPONENTS bit is set"
	(flags bitAnd: 32) = 32] whileTrue.
	hasInstr ifTrue:[
		iLen := fontFile nextNumber: 2.
		fontFile skip: iLen].! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 7/29/2009 21:59'!
processFontHeaderTable: fontFile
"Value				Data Type    Description
unitsPerEm			USHORT      Granularity of the font's em square.
xMax				USHORT      Maximum X-coordinate for the entire font.
xMin				USHORT      Minimum X-coordinate for the entire font.
yMax				USHORT      Maximum Y-coordinate for the entire font.
yMin				USHORT      Minimum Y-coordinate for the entire font.
indexToLocFormat	SHORT       Used when processing the Index To Loc Table."

	fontFile skip: 4. "Skip table version number"
	fontFile skip: 4. "Skip font revision number"
	fontFile skip: 4. "Skip check sum adjustment"
	fontFile skip: 4. "Skip magic number"
	fontFile skip: 2. "Skip flags"

	unitsPerEm := fontFile nextNumber: 2.

	fontFile skip: 8. "Skip creation date"
	fontFile skip: 8. "Skip modification date"

	"Skip min/max values of all glyphs"
	fontFile skip: 2.
	fontFile skip: 2.
	fontFile skip: 2.
	fontFile skip: 2.

	fontFile skip: 2. "Skip mac style"
	fontFile skip: 2. "Skip lowest rec PPEM"
	fontFile skip: 2. "Skip font direction hint"

	indexToLocFormat := fontFile nextNumber: 2.
! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 7/29/2009 23:45'!
processHorizontalHeaderTable: fontFile
"
ascender           SHORT          Typographic ascent.
descender          SHORT          Typographic descent.
lineGap            SHORT          Typographic lineGap.
numberOfHMetrics   USHORT         Number hMetric entries in the HTMX
                                               Table; may be smaller than the total
                                             number of glyphs.
"
	fontFile skip: 4. "Skip table version"
	ascender := self short: (fontFile nextNumber: 2).
	descender := self short: (fontFile nextNumber: 2).
	lineGap := self short: (fontFile nextNumber: 2).
	fontFile skip: 2. "Skip advanceWidthMax"
	fontFile skip: 2. "Skip minLeftSideBearing"
	fontFile skip: 2. "Skip minRightSideBearing"
	fontFile skip: 2. "Skip xMaxExtent"
	fontFile skip: 2. "Skip caretSlopeRise"
	fontFile skip: 2. "Skip caretSlopeRun"
	fontFile skip: 10. "Skip 5 reserved shorts"
	fontFile skip: 2. "Skip metricDataFormat"

	numHMetrics := fontFile nextNumber: 2.

	^numHMetrics! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 7/29/2009 21:05'!
processMaximumProfileTable: fontFile
"
numGlyphs         USHORT      The number of glyphs in the font.
"
	fontFile skip: 4. "Skip Table version number"
	numGlyphs := fontFile nextNumber: 2.! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'topa 3/9/2015 22:32'!
processNamingTable: fontFile
"copyright         CHARPTR     The font's copyright notice.
familyName        CHARPTR     The font's family name.
subfamilyName     CHARPTR     The font's subfamily name.
uniqueName        CHARPTR     A unique identifier for this font.
fullName          CHARPTR     The font's full name (a combination of
                                          familyName and subfamilyName).
versionName       CHARPTR     The font's version string.
"
	| nRecords initialOffset storageOffset format |
	initialOffset := fontFile position.
	format := fontFile nextNumber: 2.
	format = 0 ifFalse: [self error: 'Cannot handle format 1 naming tables'].
	"Get the number of name records"
	nRecords := fontFile nextNumber: 2.
	"Offset from the beginning of this table"
	storageOffset := (fontFile nextNumber: 2) + initialOffset.
	1 to: nRecords do:[:i| |  pID sID lID nID length offset string |
		fontFile position: initialOffset + 6 + ((i-1) * 12).
		pID := fontFile nextNumber: 2.
		sID := fontFile nextNumber: 2.
		lID := fontFile nextNumber: 2.
		nID := fontFile nextNumber: 2.
		length := fontFile nextNumber: 2.
		offset := fontFile nextNumber: 2.
		"Read only Macintosh or Microsoft strings"
		(pID = 1 or:[pID = 3 and:[sID = 1]]) ifTrue:[
			"MS uses Unicode all others single byte"
			"multiBytes := pID = 3."
			fontFile position: storageOffset+offset.
			string := (fontFile next: length) asString.
			pID = 3 ifTrue:[ | keep |
				keep := true.
				string := string select:[:ch| keep := keep not].
			].
			"Select only English names, prefer Macintosh"
			((pID = 1 and: [lID = 0]) or: [pID = 3 and: [lID = 16r0409]]) ifTrue: [
				nID caseOf: {
					[0] -> [(pID = 1 or:[copyright == nil]) ifTrue:[copyright := string]].
					[1] -> [(pID = 1 or:[familyName == nil]) ifTrue:[familyName := string]].
					[2] -> [(pID = 1 or:[subfamilyName == nil]) ifTrue:[subfamilyName := string]].
					"[3] -> [(pID = 1 or:[uniqueName == nil]) ifTrue:[uniqueName := string]]."
					"[4] -> [(pID = 1 or:[fullName == nil]) ifTrue:[fullName := string]]."
					"[5] -> [(pID = 1 or:[versionName == nil]) ifTrue:[versionName := string]]."
					"[6] -> [(pID = 1 or:[postscriptName == ni]) ifTrue:[postscriptName := string]]."
					"[7] -> [(pID = 1 or:[trademark == nil]) ifTrue:[trademark := string]]."
				} otherwise:["ignore"].
			]
		].
	].
! !

!TTFileDescription methodsFor: 'ttf tables' stamp: 'ar 8/29/2009 11:53'!
processOS2Table: fontFile
"
	USHORT  	 version   	0x0004
	SHORT 	xAvgCharWidth 	 
	USHORT 	usWeightClass 	 
	USHORT 	usWidthClass 	 
	USHORT 	fsType 	 
	SHORT 	ySubscriptXSize 	 
	SHORT 	ySubscriptYSize 	 
	SHORT 	ySubscriptXOffset 	 
	SHORT 	ySubscriptYOffset 	 
	SHORT 	ySuperscriptXSize 	 
	SHORT 	ySuperscriptYSize 	 
	SHORT 	ySuperscriptXOffset 	 
	SHORT 	ySuperscriptYOffset 	 
	SHORT 	yStrikeoutSize 	 
	SHORT 	yStrikeoutPosition 	 
	SHORT 	sFamilyClass 	 
	BYTE 	panose[10] 	 
	ULONG 	ulUnicodeRange1 	Bits 0-31
	ULONG 	ulUnicodeRange2 	Bits 32-63
	ULONG 	ulUnicodeRange3 	Bits 64-95
	ULONG 	ulUnicodeRange4 	Bits 96-127
	CHAR 	achVendID[4] 	 
	USHORT 	fsSelection 	 
	USHORT 	usFirstCharIndex 	 
	USHORT 	usLastCharIndex 	 
	SHORT 	sTypoAscender 	 
	SHORT 	sTypoDescender 	 
	SHORT 	sTypoLineGap 	 
	USHORT 	usWinAscent 	 
	USHORT 	usWinDescent 	 
	ULONG 	ulCodePageRange1 	Bits 0-31
	ULONG 	ulCodePageRange2 	Bits 32-63
	SHORT 	sxHeight 	 
	SHORT 	sCapHeight 	 
	USHORT 	usDefaultChar 	 
	USHORT 	usBreakChar 	 
	USHORT 	usMaxContext 	 "
	| version fsSelection minAscii maxAscii |
	version := self short: (fontFile nextNumber: 2). "table version"
	version = 0 ifTrue:[^self].
	fontFile skip: 60.
	fsSelection := fontFile nextNumber: 2.
	minAscii := fontFile nextNumber: 2.
	maxAscii := fontFile nextNumber: 2.
	sTypoAscender := self short: (fontFile nextNumber: 2).
	sTypoDescender := self short: (fontFile nextNumber: 2).
	sTypoLineGap := self short: (fontFile nextNumber: 2).
! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/30/2009 17:37'!
processSimpleGlyph: glyph contours: nContours from: fontFile
	"Construct a simple glyph frm the font file"
	| endPts  nPts iLength flags |
	endPts := Array new: nContours.
	1 to: nContours do:[:i| endPts at: i put: (fontFile nextNumber: 2)].
	glyph initializeContours: nContours with: endPts.
	nContours = 0 ifTrue:[^self].
	nPts := endPts last + 1.
	iLength := fontFile nextNumber: 2. "instruction length"
	fontFile skip: iLength.
	flags := self getGlyphFlagsFrom: fontFile size: nPts.
	self readGlyphXCoords: fontFile glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	self readGlyphYCoords: fontFile glyph: glyph nContours: nContours flags: flags endPoints: endPts.! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'nice 12/27/2009 03:12'!
readCmapTableAt: codePoint fromFile: fontFile
	| cmapFmt length firstCode entryCount segCount segIndex startCode idDelta idRangeOffset offset |
	fontFile position: cmapOffset.
	cmapFmt := fontFile nextNumber: 2.
	length := fontFile nextNumber: 2.
	fontFile skip: 2. "skip version"

	cmapFmt = 0 ifTrue:["byte encoded table"
		codePoint > 255 ifTrue:[^0].
		length := length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^0].	"but sometimes, this table is empty"
		fontFile skip: codePoint.		"move to correct byte offset in table"
		^fontFile next].

	cmapFmt = 4 ifTrue:[ "segment mapping to deltavalues"
		codePoint > 16rFFFF ifTrue:[^0].
		segCount := (fontFile nextNumber: 2) // 2.
		fontFile skip: 6. "skip searchRange, entrySelector, rangeShift"
		segIndex := (0 to: segCount-1) 
			detect:[:i| | endCode | (endCode := (fontFile nextNumber: 2)) >= codePoint].
		fontFile position: cmapOffset + 16 + (segCount*2) + (segIndex*2).
		startCode := fontFile nextNumber: 2.
		startCode <= codePoint ifFalse:[^0]. "not in segment range"
		fontFile position: cmapOffset + 16 + (segCount*4) + (segIndex*2).
		idDelta := fontFile nextNumber: 2.
		fontFile position: cmapOffset + 16 + (segCount*6) + (segIndex*2).
		idRangeOffset := fontFile nextNumber: 2.
		idRangeOffset = 0 ifTrue:[^(idDelta + codePoint) bitAnd: 16rFFFF].
		offset := (fontFile position - 2) + idRangeOffset + ((codePoint - startCode) * 2).
		fontFile position: offset.
		^fontFile nextNumber: 2.
	].

	cmapFmt = 6 ifTrue:[ "trimmed table"
		firstCode := fontFile nextNumber: 2.
		entryCount := fontFile nextNumber: 2.
		(codePoint between: firstCode and: firstCode+entryCount) ifFalse:[^0].
		fontFile skip: (codePoint-firstCode) * 2.
		^fontFile nextNumber: 2].

	^0! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/30/2009 17:42'!
readGlyphAt: glyphIndex
	| glyph |
	self withFileDo:[:fontFile|
		glyph := self readGlyphAt: glyphIndex fromFile: fontFile.
		self updateGlyphMetrics: glyph fromFile: fontFile.
	].
	^glyph! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/30/2009 17:06'!
readGlyphAt: glyphIndex fromFile: fontFile
	"Answer the glyph with the given glyph index"
	| glyphOffset nextOffset glyphLength glyph nContours left top right bottom |
	indexToLocFormat = 0 ifTrue:["Format0: offset/2 is stored"
		fontFile position: indexToLocOffset+(glyphIndex * 2).
		glyphOffset := (fontFile nextNumber: 2) * 2.
		nextOffset := (fontFile nextNumber: 2) * 2.
	] ifFalse:["Format1: store actual offset"
		fontFile position: indexToLocOffset+(glyphIndex * 4).
		glyphOffset := fontFile nextNumber: 4.
		nextOffset := fontFile nextNumber: 4.
	].
	glyphLength := nextOffset - glyphOffset.
	glyphLength = 0 ifTrue:[^TTGlyph new glyphIndex: glyphIndex].

	fontFile position: glyphTableOffset+glyphOffset.
	nContours := self short: (fontFile nextNumber: 2).
	left := self short: (fontFile nextNumber: 2).
	top := self short: (fontFile nextNumber: 2).
	right := self short: (fontFile nextNumber: 2).
	bottom := self short: (fontFile nextNumber: 2).
	nContours >= 0 ifTrue:[
		glyph := TTGlyph new glyphIndex: glyphIndex.
		self processSimpleGlyph: glyph contours: nContours from: fontFile.
	] ifFalse:[
		glyph := TTCompositeGlyph new glyphIndex: glyphIndex.
		self processCompositeGlyph: glyph contours: nContours from: fontFile.
	].
	glyph buildAllContours.
	glyph bounds: (left@top corner: right@bottom).
	^glyph
! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/29/2009 21:42'!
readGlyphXCoords: fontFile glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the x coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits xValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"If bit zero in the flag is set then this point is an on-curve
			point, if not, then it is an off-curve point."
			(flagBits bitAnd: 1) = 1 
				ifTrue:[ ttPoint type: #OnCurve]
				ifFalse:[ttPoint type: #OffCurve].
			"First we check to see if bit one is set.  This would indicate that
			the corresponding coordinate data in the table is 1 byte long.
			If the bit is not set, then the coordinate data is 2 bytes long."
			(flagBits bitAnd: 2) = 2 ifTrue:[ "one byte"
				xValue := fontFile next.
				xValue := (flagBits bitAnd: 16)=16 ifTrue:[xValue] ifFalse:[xValue negated].
				ttPoint x: xValue.
			] ifFalse:[ "two byte"
				"If bit four is set, then this coordinate is the same as the
				last one, so the relative offset (of zero) is stored.  If bit
				is not set, then read in two bytes and store it as a signed value."
				(flagBits bitAnd: 16) = 16 ifTrue:[ ttPoint x: 0 ]
				ifFalse:[
					xValue := self short: (fontFile nextNumber: 2).
					ttPoint x: xValue]]].
		startPoint := endPoint + 1]! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/29/2009 21:44'!
readGlyphYCoords: fontFile glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the y coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits yValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"Check if this value one or two byte encoded"
			(flagBits bitAnd: 4) = 4 ifTrue:[ "one byte"
				yValue := fontFile next.
				yValue := (flagBits bitAnd: 32)=32 ifTrue:[yValue] ifFalse:[yValue negated].
				ttPoint y: yValue.
			] ifFalse:[ "two byte"
				(flagBits bitAnd: 32) = 32 ifTrue:[ ttPoint y: 0 ]
				ifFalse:[
					yValue := self short: (fontFile nextNumber: 2).
					ttPoint y: yValue]]].
		startPoint := endPoint + 1]! !

!TTFileDescription methodsFor: 'rendering' stamp: 'topa 3/9/2015 23:26'!
renderFallbackGlyphOfHeight: height fgColor: fgColor bgColor: bgColor depth: depth
	"Render the glyph with the given code point at the specified pixel height."
	^ self fallbackGlyph
		asFormWithScale: height asFloat / (ascender - descender) 
			ascender: ascender 
			descender: descender 
			fgColor: fgColor bgColor: bgColor depth: depth! !

!TTFileDescription methodsFor: 'rendering' stamp: 'ar 8/29/2009 19:11'!
renderGlyph: code height: height fgColor: fgColor bgColor: bgColor depth: depth
	"Render the glyph with the given code point at the specified pixel height."
	^(self at: code) 
		asFormWithScale: height asFloat / (ascender - descender) 
			ascender: ascender 
			descender: descender 
			fgColor: fgColor bgColor: bgColor depth: depth! !

!TTFileDescription methodsFor: 'private' stamp: 'ar 7/29/2009 21:25'!
short: aNumber
	(aNumber bitAnd: 16r8000) = 0
		ifTrue: [^aNumber]
		ifFalse: [^-1 - (aNumber bitXor: 16rFFFF)]! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/30/2009 00:16'!
size
	"Compatibility with TTFontDescription"
	^16rFFFF! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 7/29/2009 23:52'!
subfamilyName
	"The subfamily name for the font"
	^subfamilyName! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicAscender
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoAscender ifNil:[ascender]! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicDescender
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoDescender ifNil:[descender]! !

!TTFileDescription methodsFor: 'accessing' stamp: 'topa 4/3/2015 19:48'!
typographicFontHeight
	"cope for the fact that typographicAscender and 
	typographicDescender may not be available and 
	0-height fonts are a bit useless"
	| tfh |
	tfh := self typographicAscender - self typographicDescender.
	^ tfh = 0 ifTrue: [self fontHeight] ifFalse: [tfh]! !

!TTFileDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicLineGap
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoLineGap ifNil:[lineGap]! !

!TTFileDescription methodsFor: 'glyphs' stamp: 'ar 7/29/2009 23:51'!
updateGlyphMetrics: glyph fromFile: fontFile
	"Update the horizontal metrics for the given glyph"
	| glyphIndex |
	glyphIndex := glyph glyphIndex.
	glyphIndex <= numHMetrics ifTrue:[
		fontFile position: hmtxTableOffset + (glyphIndex*4).
		glyph advanceWidth: (fontFile nextNumber: 2).
		glyph leftSideBearing: (self short: (fontFile nextNumber: 2)).
	] ifFalse:[
		fontFile position: hmtxTableOffset + ((numHMetrics-1) *4).
		glyph advanceWidth: (fontFile nextNumber: 2).
		fontFile position: hmtxTableOffset + (numHMetrics * 4) + ((glyphIndex-numHMetrics)*2).
		glyph leftSideBearing: (self short: (fontFile nextNumber: 2)).
	].
	glyph updateRightSideBearing.! !

!TTFileDescription methodsFor: 'initialize' stamp: 'ar 7/30/2009 19:54'!
withFileDo: aBlock
	"Open the font file for the duration of aBlock"
	^self class openFontFile: fileName do: aBlock.! !

Object subclass: #TTFontDescription
	instanceVariableNames: 'glyphTable glyphs kernPairs copyright familyName fullName subfamilyName uniqueName versionName postscriptName trademark bounds unitsPerEm ascender descender lineGap sTypoAscender sTypoDescender sTypoLineGap'
	classVariableNames: 'Default Descriptions'
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTFontDescription commentStamp: '<historical>' prior: 0!
Holds a TrueType font in memory.  Is used by TTSampleStringMorph as its font.  

Class owns a default example.  !

TTFontDescription subclass: #TTCFontDescription
	instanceVariableNames: ''
	classVariableNames: 'TTCDefault TTCDescriptions'
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCFontDescription class methodsFor: 'instance creations' stamp: 'topa 3/19/2015 11:17'!
addFromTTFile: fileName
"
	Execute the following only if you know what you are doing.
	self addFromTTFile: 'C:\WINDOWS\Fonts\msgothic.TTC'
"

	| tt |
	(fileName asLowercase endsWith: 'ttf') ifTrue: [
		tt := TTCFontReader readTTFFrom: (FileStream readOnlyFileNamed: fileName).
	] ifFalse: [
		tt := TTCFontReader readFrom: (FileStream readOnlyFileNamed: fileName).
	].

	self addToDescription: tt.
	^ tt.
! !

!TTCFontDescription class methodsFor: 'accessing' stamp: 'tk 12/6/2004 09:05'!
descriptionNamed: descriptionName at: index

	| array |
	(array :=  self descriptionNamed: descriptionName) ifNil: [^ nil].
	^ array at: index.
! !

!TTCFontDescription methodsFor: 'accessing' stamp: 'yo 12/27/2002 04:25'!
at: aCharOrInteger

	| char |
	char := aCharOrInteger asCharacter.
	^ glyphs at: (char charCode) + 1.
! !

!TTCFontDescription methodsFor: 'copying' stamp: 'yo 11/14/2002 22:40'!
deepCopy

	^ self.
! !

!TTCFontDescription methodsFor: 'objects from disk' stamp: 'yo 11/30/2002 22:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to a known Font in the other system instead.  "

	"A path to me"
	(TextConstants at: #forceFontWriting ifAbsent: [false]) ifTrue: [^ self].
		"special case for saving the default fonts on the disk.  See collectionFromFileNamed:"

	dp := DiskProxy global: #TTCFontDescription selector: #descriptionNamed:at:
			args: {self name. ((TTCFontDescription descriptionNamed: self name) indexOf: self)}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTCFontDescription methodsFor: 'accessing' stamp: 'yo 11/14/2002 18:28'!
size

	^ glyphs size.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 12/11/2007 22:41'!
addFromSmartRefStream: ref

	| tts |
	tts := ref nextAndClose.

	^ tts collect: [:tt |
		self addToDescription: tt.
	].
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'tb 6/24/2003 17:10'!
addFromTTFile: fileName
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"
	^self addFromTTStream: (FileStream readOnlyFileNamed: fileName).
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/10/2015 19:28'!
addFromTTStream: readStream
"
	self addFromTTFile: 'C:\WINDOWS\Fonts\ARIALN.TTF'
"

	| tt |
	tt := TTFontReader readFrom: readStream.
	tt := self addToDescription: tt.
	tt blankGlyphForSeparators.
	^ tt.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/10/2015 19:33'!
addSetFromTTFile: fileName encodingTag: encodingTag ranges: ranges

	| tt |
	(fileName asLowercase endsWith: 'ttf') ifTrue: [
		tt := TTCFontReader readTTFFrom: (FileStream readOnlyFileNamed: fileName).
	] ifFalse: [
		tt := TTCFontReader readFrom: (FileStream readOnlyFileNamed: fileName).
	].

	(tt at: encodingTag + 1) compactForRanges: ranges.
	^ self addToDescription: tt.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/10/2015 19:29'!
addToDescription: tt

	| old new |
	old := Descriptions detect: [:f | f first fullName = tt first fullName] ifNone: [nil].
	^ old ifNotNil: [
		new := old, (Array new: ((tt size - old size) max: 0)).
		1 to: tt size do: [:ind |
			(tt at: ind) ifNotNil: [
				new at: ind put: (tt at: ind)
			].
		].
		Descriptions remove: old.
		Descriptions add: new.
		new.
	] ifNil: [
		Descriptions add: tt.
		tt.
	]
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 11/30/2002 22:22'!
clearDefault
"
	self clearDefault
"

	Default := nil.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 11/30/2002 22:22'!
clearDescriptions
"
	self clearDescriptions
"

	Descriptions := Set new.
	Default ifNotNil: [Descriptions add: Default].
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'tk 12/10/2001 17:12'!
default
	^ Default! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 12/7/2007 23:51'!
descriptionFullNamed: descriptionFullName 
	^ Descriptions
		detect: [:f | f first fullName = descriptionFullName]
		ifNone: [Default]! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/10/2015 19:31'!
descriptionFullNamed: descriptionFullName at: index
	| ans |
	ans := Descriptions
		detect: [:f | f first fullName = descriptionFullName]
		ifNone: [Default].
	index > 0 ifTrue: [^ ans at: index].
	^ ans.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 12/7/2007 13:26'!
descriptionNamed: descriptionName

	^ Descriptions detect: [:f | f first name = descriptionName] ifNone: [Default].
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/19/2015 11:19'!
foo! !

!TTFontDescription class methodsFor: 'class initialization' stamp: 'yo 11/30/2002 22:22'!
initialize
"
	self initialize
"

	self clearDescriptions.
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'topa 3/10/2015 19:31'!
removeDescriptionNamed: descriptionName

	| tt |
	Descriptions ifNil: [^ self].
	[(tt :=  Descriptions detect: [:f | f first name = descriptionName] ifNone: [nil]) notNil] whileTrue:[
		 Descriptions remove: tt
	].
! !

!TTFontDescription class methodsFor: 'instance creations' stamp: 'yo 12/7/2007 23:51'!
removeDescriptionNamed: descriptionName subfamilyName: subfamilyName

	| tts |
	Descriptions ifNil: [^ self].
	tts _ Descriptions select: [:f | f first name = descriptionName and: [f first subfamilyName = subfamilyName]].
	tts do: [:f | Descriptions remove: f].
! !

!TTFontDescription class methodsFor: 'instance creation' stamp: 'yo 12/13/2002 13:20'!
setDefault
"
	self setDefault
"

	Default := TTFontReader readFrom: (FileStream readOnlyFileNamed: 'C:\WINDOWS\Fonts\comic.ttf').
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 8/29/2009 19:10'!
ascender
	"Ascender of the font. Relative to unitsPerEm.
	Easily confused with the typographic ascender."
	^ascender! !

!TTFontDescription methodsFor: 'converting' stamp: 'sma 5/5/2000 13:46'!
asStrikeFontScale: scale
	"Generate a StrikeFont (actually a FormSetFont) for this TTF font at a given scale."

	| forms |
	forms := (0 to: 255) collect:
		[:i |
		(self at: i)
			asFormWithScale: scale
			ascender: ascender
			descender: descender].
	^ FormSetFont new
		fromFormArray: forms
		asciiStart: 0
		ascent: (ascender * scale) rounded! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 12/6/2007 14:52'!
at: aCharOrInteger
	^glyphTable at: (aCharOrInteger isCharacter ifTrue: [aCharOrInteger charCode] ifFalse: [aCharOrInteger])+1! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:08'!
at: index put: value
	^self shouldNotImplement! !

!TTFontDescription methodsFor: 'migration' stamp: 'yo 8/16/2004 10:44'!
blankGlyphForSeparators

	| space |
	space := (self at: Character space charCode) copy.
	space contours: #().
	Character separators do: [:s | 
		glyphTable at: s charCode +1 put: space.
	].
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:48'!
bounds
	^bounds! !

!TTFontDescription methodsFor: 'migration' stamp: 'topa 3/10/2015 19:35'!
compactForRanges: rangesArray

	| newGlyphTable noMapping |
	noMapping := glyphs == glyphTable.
	newGlyphTable :=  SparseLargeTable new: rangesArray last last chunkSize: 32 arrayClass: Array base: 0 + 1 defaultValue: (glyphs at: 1).
	rangesArray do: [:pair |
		pair first to: pair second do: [:i |
			newGlyphTable at: i put: (glyphs at: i)
		]
	].
	glyphTable := newGlyphTable.
	noMapping ifTrue: [glyphs := glyphTable].
! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
copyright
	^copyright! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 6/23/2003 18:23'!
deepCopy

	"Since it shouldn't be copied for transmitting or any reason, it returns self."
	^ self.
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 8/29/2009 19:11'!
descender
	"Descender of the font. Relative to unitsPerEm.
	Easily confused with the typographic descender."
	^descender! !

!TTFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 23:00'!
fallbackGlyph
	"Answer the fallback glyph, the first in the loca table "
	^ glyphs at: 1! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
familyName
	^familyName! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 12/7/2007 10:33'!
first

	^ self.
! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/14/1998 20:20'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	glyphs do:[:glyph| glyph flipAroundY]! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 18:46'!
fontHeight
	^ascender - descender! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
fullName
	^fullName! !

!TTFontDescription methodsFor: 'testing' stamp: 'topa 3/19/2015 03:07'!
isRegular
	"Answer true if I am a Regular/Roman font (i.e. not bold, etc.)"
	^ (TTCFont indexOfSubfamilyName: (self subfamilyName)) = 0.
! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 8/29/2009 19:11'!
lineGap
	"Leading of the font. Relative to unitsPerEm.
	Easily confused with the typographic linegap."
	^lineGap! !

!TTFontDescription methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:38'!
name

	^ self familyName copyWithout: Character space.
! !

!TTFontDescription methodsFor: 'copying' stamp: 'topa 3/10/2015 19:36'!
objectForDataStream: refStrm
	| dp isCollection |
	"I am about to be written on an object file.  Write a reference to a known Font in the other system instead.  "

	"A path to me"
	(TextConstants at: #forceFontWriting ifAbsent: [false]) ifTrue: [^ self].
		"special case for saving the default fonts on the disk.  See collectionFromFileNamed:"

	isCollection := (Descriptions detect: [:e | e == self]) isCollection.

	dp := DiskProxy global: #TTFontDescription selector: #descriptionFullNamed:at:
			args: {self fullName. (isCollection ifTrue: [(TTFontDescription descriptionFullNamed: self fullName) indexOf: self] ifFalse: [0])}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
postscriptName
	^postscriptName! !

!TTFontDescription methodsFor: 'printing' stamp: 'th 6/27/2003 17:08'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	familyName printOn: aStream.
	aStream nextPut:$).! !

!TTFontDescription methodsFor: 'rendering' stamp: 'topa 3/9/2015 23:27'!
renderFallbackGlyphOfHeight: height fgColor: fgColor bgColor: bgColor depth: depth
	"Render the glyph with the given code point at the specified pixel height."
	^ self fallbackGlyph
		asFormWithScale: height asFloat / (ascender - descender) 
			ascender: ascender 
			descender: descender 
			fgColor: fgColor bgColor: bgColor depth: depth! !

!TTFontDescription methodsFor: 'rendering' stamp: 'ar 8/29/2009 19:12'!
renderGlyph: code height: fontHeight fgColor: fgColor bgColor: bgColor depth: depth
	"Render the glyph with the given code point at the specified pixel height."
	^(self at: code) 
		asFormWithScale: fontHeight asFloat / (ascender - descender) 
			ascender: ascender 
			descender: descender 
			fgColor: fgColor bgColor: bgColor depth: depth! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setAscender: asc descender: desc lineGap: lgap
	ascender := asc.
	descender := desc.
	lineGap := lgap! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:28'!
setBounds: aRect unitsPerEm: aNumber
	bounds := aRect.
	unitsPerEm := aNumber.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:27'!
setGlyphs: glyphArray mapping: mappingTable
	glyphs := glyphArray.
	glyphTable := mappingTable.! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:48'!
setKernPairs: array
	kernPairs := array! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 11/2/1998 00:46'!
setStrings: anArray
	copyright := anArray at: 1.
	familyName := anArray at: 2.
	subfamilyName := anArray at: 3.
	uniqueName := anArray at: 4.
	fullName := anArray at: 5.
	versionName := anArray at: 6.
	postscriptName := anArray at: 7.
	trademark := anArray at: 8.
! !

!TTFontDescription methodsFor: 'private-initialization' stamp: 'ar 8/29/2009 19:19'!
setTypographicAscender: asc descender: desc lineGap: lGap
	sTypoAscender := asc.
	sTypoDescender := desc.
	sTypoLineGap := lGap.
! !

!TTFontDescription methodsFor: 'accessing' stamp: 'bf 9/29/2010 11:09'!
size

	^ glyphTable size.
! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
subfamilyName
	^subfamilyName! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
trademark
	^trademark! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicAscender
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoAscender ifNil:[ascender]! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicDescender
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoDescender ifNil:[descender]! !

!TTFontDescription methodsFor: 'accessing' stamp: 'topa 4/3/2015 21:56'!
typographicFontHeight
	"cope for the fact that typographicAscender and 
	typographicDescender may not be available and 
	0-height fonts are a bit useless"
	| tfh |
	tfh := self typographicAscender - self typographicDescender.
	^ tfh = 0 ifTrue: [self fontHeight] ifFalse: [tfh]! !

!TTFontDescription methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:10'!
typographicLineGap
	"Microsoft defines this as the 'true typographic metrics' of the font."
	^sTypoLineGap ifNil:[lineGap]! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:48'!
uniqueName
	^uniqueName! !

!TTFontDescription methodsFor: 'properties' stamp: 'ar 11/14/1998 23:49'!
unitsPerEm
	^unitsPerEm! !

!TTFontDescription methodsFor: 'information' stamp: 'ar 11/14/1998 23:49'!
versionName
	^versionName! !

!TTFontDescription methodsFor: 'copying' stamp: 'yo 11/30/2002 22:38'!
veryDeepCopyWith: deepCopier
	"Return self.  I am shared.  Do not record me."
! !

Object subclass: #TTFontReader
	instanceVariableNames: 'charMap glyphs nGlyphs kernPairs infoBar fontDescription'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTFontReader commentStamp: '<historical>' prior: 0!
TTFontReader constructs a TTFontDescription from a TrueType font (.ttf).!

TTFontReader subclass: #TTCFontReader
	instanceVariableNames: 'fonts'
	classVariableNames: 'EncodingTag'
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCFontReader class methodsFor: 'as yet unclassified' stamp: 'yo 2/15/2004 18:45'!
encodingTag: aNumber
"
	TTCFontReader encodingTag: 6
"

	EncodingTag := aNumber.
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'yo 11/8/2002 15:12'!
getTableDirEntry: tagString from: fontData offset: offset
	"Find the table named tagString in fontData and return a table directory entry for it."

	| nTables pos currentTag tag |
	nTables := fontData shortAt: 5 + offset bigEndian: true.
	tag := ByteArray new: 4.
	1 to: 4 do:[:i| tag byteAt: i put: (tagString at: i) asInteger].
	tag := tag longAt: 1 bigEndian: true.
	pos := 13 + offset.
	1 to: nTables do:[:i|
		currentTag := fontData longAt: pos bigEndian: true.
		currentTag = tag ifTrue:[^TTFontTableDirEntry on: fontData at: pos].
		pos := pos+16].
	^nil! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'yo 11/8/2002 15:47'!
parseTTCHeaderFrom: fontData

	| pos nTables |
	nTables := fontData longAt: 9 bigEndian: true.
	fonts := Array new: nTables.
	pos := 13.
	1 to: nTables do: [:i |
		fonts at: i put: (fontData longAt: pos bigEndian: true).
		pos := pos + 4.
	].

	^ fonts
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'vv 8/29/2015 00:52'!
processCharMap: assoc
	"Process the given character map"

	| glyph cmap encode0 encode1 char value null |
	cmap := assoc value.
	null := (glyphs at: (cmap at: Character space asUnicode + 1) + 1) copy.
	null contours: #().

	encode0 := Array new: 256 withAll: glyphs first.
	encode1 := Array new: 65536 withAll: glyphs first.

	0 to: 255 do: [:i |
		char := Character value: i.
		glyph := glyphs at: (cmap at: char asUnicode + 1) + 1.
		encode0 at: i+1 put: glyph.
	].
	Character separators do: [:c |
		encode0 at: (c asciiValue + 1) put: null.
	].
	0 to: 65536 - 1 do: [:i |
		value := cmap at: i+1.
		value = 65535 ifFalse: [ "???"
			| g |
			g := glyphs at: value+1 ifAbsent: [ null. ].
			(g isKindOf: TTCompositeGlyph) ifFalse: [
				encode1 at: i+1 put: g.
			] ifTrue: [
				g basicGlyphs: (((glyphs at: value+1) basicGlyphs) collect: [:t | t key->(glyphs at: (t value glyphIndex+1))]).
				encode1 at: i+1 put: g
			].
		]
	].

	^ {encode0. encode1}.
! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'topa 3/19/2015 02:33'!
readFrom: aStream

	"Read the raw font byte data"
	| fontData |
	(aStream respondsTo: #binary) ifTrue:[aStream binary].
	fontData := aStream contents asByteArray.

	fonts := self parseTTCHeaderFrom: fontData.
	^ fonts gather: [:offset |
		fontDescription := TTCFontDescription new.
		self readFrom: fontData fromOffset: offset at: EncodingTag]! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'topa 3/19/2015 02:09'!
readFrom: fontData fromOffset: offset at: encodingTag

	| headerEntry maxProfileEntry nameEntry indexLocEntry charMapEntry glyphEntry horzHeaderEntry horzMetricsEntry kerningEntry glyphOffset cmap numHMetrics indexToLocFormat fontDescription0 fontDescription1 array result |

	"Search the tables required to build the font"
	(headerEntry := self getTableDirEntry: 'head' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a header table'].
	(maxProfileEntry := self getTableDirEntry: 'maxp' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a maximum profile table'].
	(nameEntry := self getTableDirEntry: 'name' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a name table'].
	(indexLocEntry := self getTableDirEntry: 'loca' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a relocation table'].
	(charMapEntry := self getTableDirEntry: 'cmap' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a character map table'].
	(glyphEntry := self getTableDirEntry: 'glyf' from: fontData  offset: offset) == nil ifTrue:[
		^self error:'This font does not have a glyph table'].
	(horzHeaderEntry := self getTableDirEntry: 'hhea' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a horizontal header table'].
	(horzMetricsEntry := self getTableDirEntry: 'hmtx' from: fontData offset: offset) == nil ifTrue:[
		^self error:'This font does not have a horizontal metrics table'].
	(kerningEntry := self getTableDirEntry: 'kern' from: fontData offset: offset) == nil ifTrue:[
		Transcript cr; show:'This font does not have a kerning table';endEntry].


	"Process the data"
	indexToLocFormat := self processFontHeaderTable: headerEntry.
	self processMaximumProfileTable: maxProfileEntry.
	self processNamingTable: nameEntry.
	glyphOffset := self processIndexToLocationTable: indexLocEntry format: indexToLocFormat.
	cmap := self processCharacterMappingTable: charMapEntry.
	(cmap == nil or:[cmap value == nil])
		ifTrue:[^self error:'This font has no suitable character mappings'].
	self processGlyphDataTable: glyphEntry offsets: glyphOffset.
	numHMetrics := self processHorizontalHeaderTable: horzHeaderEntry.
	self processHorizontalMetricsTable: horzMetricsEntry length: numHMetrics.
	kerningEntry isNil 
		ifTrue:[kernPairs := #()]
		ifFalse:[self processKerningTable: kerningEntry].
	array := self processCharMap: cmap.
	fontDescription0 := fontDescription clone.
	fontDescription1 := fontDescription clone.
	fontDescription0 setGlyphs: (array at: 1) mapping: (array at: 1)..
	fontDescription1 setGlyphs: (array at: 2) mapping: (array at: 2)..
	fontDescription0 setKernPairs: kernPairs.
	fontDescription1 setKernPairs: kernPairs.
	result := OrderedCollection new.
	(encodingTag = nil or: [encodingTag = 0]) ifTrue: [^ Array with: fontDescription1].
	result add: fontDescription0.
	encodingTag -1 timesRepeat: [result add: nil].
	result add: fontDescription1.
	^ result asArray.

! !

!TTCFontReader methodsFor: 'as yet unclassified' stamp: 'yo 2/15/2004 18:43'!
readTTFFrom: aStream

	"Read the raw font byte data"
	| fontData |
	(aStream respondsTo: #binary) ifTrue:[aStream binary].
	fontData := aStream contents asByteArray.
	fontDescription := TTCFontDescription new.

	^ self readFrom: fontData fromOffset: 0 at: EncodingTag.
! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'nk 7/16/2003 15:56'!
fileReaderServicesForFile: fullName suffix: suffix


	^(suffix = 'fnt')  | (suffix = '*') 
		ifTrue: [ self services]
		ifFalse: [#()]
! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'ar 9/5/2010 09:38'!
initialize
	"self initialize"

	FileServices registerFileReader: self! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'topa 3/18/2015 20:37'!
installTTF: ttfFileName asTextStyle: textStyleName sizes: sizeArray
	"Sizes are in pixels."
	"TTFontReader
		installTTF: 'F:\fonts\amazon__.TTF' 
		asTextStyle: #Amazon
		sizes: #(24 60)"

	| ttf fontArray |
	ttf := self parseFileNamed: ttfFileName.
	fontArray := sizeArray collect:
		[:each |
		(ttf asStrikeFontScale: each / ttf unitsPerEm)
			name: textStyleName;
			pixelSize: each].
	TextConstants at: textStyleName asSymbol put: (TextStyle fontArray: fontArray)! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 19:31'!
openTTFFile: fullName 

	(TTFontReader parseFileNamed: fullName) asMorph open! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'ar 7/29/2009 21:09'!
parseFileNamed: aString
	"TTFontReader parseFileNamed:'c:\windows\fonts\arial.ttf'"
	"TTFontReader parseFileNamed:'c:\windows\times.ttf'"
	| contents |
	contents := (FileStream readOnlyFileNamed: aString) binary contentsOfEntireFile.
	^self readFrom: (ReadStream on: contents)! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 00:53'!
readFrom: aStream

	^self new readFrom: aStream! !

!TTFontReader class methodsFor: 'instance creation' stamp: 'yo 2/15/2004 18:40'!
readTTFFrom: aStream

	^self new readTTFFrom: aStream! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'KR 3/21/2008 20:52'!
serviceOpenTrueTypeFont

	^ SimpleServiceEntry 
				provider: self 
				label: 'open true type font' translatedNoop
				selector: #openTTFFile:
				description: 'open true type font' translatedNoop! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'sd 2/1/2002 22:03'!
services

	^ Array with: self serviceOpenTrueTypeFont
! !

!TTFontReader class methodsFor: 'class initialization' stamp: 'ar 9/5/2010 09:38'!
unload

	FileServices unregisterFileReader: self ! !

!TTFontReader methodsFor: 'private' stamp: 'topa 9/1/2015 20:14'!
decodeCmapFmtTable: entry
	| cmapFmt length entryCount segCount segments offset cmap firstCode |
	cmapFmt := entry nextUShort.
	length := entry nextUShort.
	entry skip: 2. "skip version"

	cmapFmt = 0 ifTrue: "byte encoded table"
		[length := length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^ nil].	"but sometimes, this table is empty"
		cmap := Array new: length.
		entry nextBytes: length into: cmap startingAt: entry offset.
		^ cmap].

	cmapFmt = 4 ifTrue: "segment mapping to deltavalues"
		[segCount := entry nextUShort // 2.
		entry skip: 6. "skip searchRange, entrySelector, rangeShift"
		segments := Array new: segCount.
		segments := (1 to: segCount) collect: [:e | Array new: 4].
		1 to: segCount do: [:i | (segments at: i) at: 2 put: entry nextUShort]. "endCount"
		entry skip: 2. "skip reservedPad"
		1 to: segCount do: [:i | (segments at: i) at: 1 put: entry nextUShort]. "startCount"
		1 to: segCount do: [:i | (segments at: i) at: 3 put: entry nextShort]. "idDelta"
		offset := entry offset.
		1 to: segCount do: [:i | (segments at: i) at: 4 put: entry nextUShort]. "idRangeOffset"
		entryCount := segments inject: 0 into: [:max :seg | max max: seg second].
		cmap := Array new: entryCount+1 withAll: 0..
		segments withIndexDo:
			[:seg :si | | code |
			seg first to: seg second do:
				[:i |
					seg last > 0 ifTrue:
						["offset to glypthIdArray - this is really C-magic!!"
						entry offset: i - seg first - 1 * 2 + seg last + si + si + offset. 
						code := entry nextUShort.
						code > 0 ifTrue: [code := code + seg third]]
					ifFalse:
						["simple offset"
						code := i + seg third].
					cmap at: i + 1 put: code]].
		^ cmap].

	cmapFmt = 6 ifTrue: "trimmed table"
		[firstCode := entry nextUShort.
		entryCount := entry nextUShort.
		cmap := Array new: entryCount + firstCode withAll: 0.
		entryCount timesRepeat:
			[cmap at: (firstCode := firstCode + 1) put: entry nextUShort].
		^ cmap].
	^ nil! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getGlyphFlagsFrom: entry size: nPts
	"Read in the flags for this glyph.  The outer loop gathers the flags that
	are actually contained in the table.  If the repeat bit is set in a flag
	then the next byte is read from the table; this is the number of times
	to repeat the last flag.  The inner loop does this, incrementing the
	outer loops index each time."
	| flags index repCount flagBits |
	flags := ByteArray new: nPts.
	index := 1.
	[index <= nPts] whileTrue:[
		flagBits := entry nextByte.
		flags at: index put: flagBits.
		(flagBits bitAnd: 8) = 8 ifTrue:[
			repCount := entry nextByte.
			repCount timesRepeat:[
				index := index + 1.
				flags at: index put: flagBits]].
		index := index + 1].
	^flags! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:33'!
getTableDirEntry: tagString from: fontData
	"Find the table named tagString in fontData and return a table directory entry for it."
	| nTables pos currentTag tag |
	nTables := fontData shortAt: 5 bigEndian: true.
	tag := ByteArray new: 4.
	1 to: 4 do:[:i| tag byteAt: i put: (tagString at: i) asInteger].
	tag := tag longAt: 1 bigEndian: true.
	pos := 13.
	1 to: nTables do:[:i|
		currentTag := fontData longAt: pos bigEndian: true.
		currentTag = tag ifTrue:[^TTFontTableDirEntry on: fontData at: pos].
		pos := pos+16].
	^nil! !

!TTFontReader methodsFor: 'private' stamp: 'ar 4/10/2005 16:10'!
macToWin: index
	^ (index - 1) asCharacter macToSqueak asciiValue + 1! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/19/2015 11:48'!
processCharacterMappingTable: entry
	"Read the font's character to glyph index mapping table.
	If an appropriate mapping can be found then return an association
	with the format identifier and the contents of the table"
	| copy initialOffset nSubTables pID sID offset cmap assoc |
	initialOffset := entry offset.
	entry skip: 2. "Skip table version"
	nSubTables := entry nextUShort.
	1 to: nSubTables do:[:i|
		pID := entry nextUShort.
		sID := entry nextUShort.
		offset := entry nextULong.
		"Check if this is either a Macintosh encoded table
		or a Windows encoded table"
		(#(0 1 3) includes: pID) ifTrue: [
			(assoc notNil and: [assoc key = pID]) ifFalse: [
				"Go to the beginning of the table"
				copy := entry copy.
				copy offset: initialOffset + offset.
				cmap := self decodeCmapFmtTable: copy.
				(pID = 0 and: [cmap notNil]) "Prefer Unicode encoding over everything else"
					ifTrue: [^ pID -> cmap].
				"(pID = 1 and: [cmap notNil])" "Prefer Macintosh encoding over everything else"
					"ifTrue: [pID -> cmap]."
				assoc := pID -> cmap. "Keep it in case we don't find a better table"
			].
		].
	].
	^assoc! !

!TTFontReader methodsFor: 'processing' stamp: 'bf 6/12/2006 18:00'!
processCharMap: assoc
	"Process the given character map"

	| charTable glyph cmap |
	cmap := assoc value.

	assoc key = 0 ifTrue: "Unicode table"
		[charTable := SparseLargeTable new: cmap size
			chunkSize: 256 arrayClass: Array base: 1
			defaultValue: glyphs first.
		1 to: charTable size do:
			[:i |
			glyph := glyphs at: (cmap at: i) + 1 ifAbsent: [glyphs first].
			charTable at: i put: glyph].
		charTable zapDefaultOnlyEntries.
		^charTable].

	charTable := Array new: 256 withAll: glyphs first. "Initialize with default glyph"

	assoc key = 1 ifTrue: "Mac encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph := glyphs at: (cmap at: i) + 1.
			charTable at: (self macToWin: i) put: glyph]].

	assoc key = 3 ifTrue: "Win encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph := glyphs at: (cmap at: i) + 1.
			charTable at: i put: glyph]].

	^ charTable! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/10/2015 19:50'!
processCompositeGlyph: glyph contours: nContours from: entry
	"Read a composite glyph from the font data. The glyph passed into this method contains some state variables that must be copied into the resulting composite glyph."
	| flags glyphIndex hasInstr cGlyph ofsX ofsY iLen a11 a12 a21 a22 m |
	cGlyph := TTCompositeGlyph new.
	a11 := a22 := 16r4000.	"1.0 in F2Dot14"
	a21 := a12 := 0.		"0.0 in F2Dot14"
	"Copy state"
	cGlyph bounds: glyph bounds; glyphIndex: glyph glyphIndex.
	hasInstr := false.
	[ flags := entry nextUShort.
	glyphIndex := entry nextUShort + 1.
	(flags bitAnd: 1) = 1 ifTrue:[
		ofsX := entry nextShort.
		ofsY := entry nextShort.
	] ifFalse:[
		(ofsX := entry nextByte) > 127 ifTrue:[ofsX := ofsX - 256].
		(ofsY := entry nextByte) > 127 ifTrue:[ofsY := ofsY - 256]].
	((flags bitAnd: 2) = 2) ifFalse:[self halt].
	(flags bitAnd: 8) = 8 ifTrue:[
		a11 := a22 := entry nextShort].
	(flags bitAnd: 64) = 64 ifTrue:[
		a11 := entry nextShort.
		a22 := entry nextShort].
	(flags bitAnd: 128) = 128 ifTrue:[
		"2x2 transformation"
		a11 := entry nextShort.
		a21 := entry nextShort.
		a12 := entry nextShort.
		a22 := entry nextShort].
	m := MatrixTransform2x3 new.
	"Convert entries from F2Dot14 to float"
	m a11: (a11 asFloat / 16r4000).
	m a12: (a12 asFloat / 16r4000).
	m a21: (a21 asFloat / 16r4000).
	m a22: (a22 asFloat / 16r4000).
	m a13: ofsX.
	m a23: ofsY.
	cGlyph addGlyph: (glyphs at: glyphIndex) transformation: m.
	hasInstr := hasInstr or:[ (flags bitAnd: 256) = 256].
	"Continue as long as the MORE_COMPONENTS bit is set"
	(flags bitAnd: 32) = 32] whileTrue.
	hasInstr ifTrue:[
		iLen := entry nextUShort.
		entry skip: iLen].
	^cGlyph! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:42'!
processFontHeaderTable: entry
"Value				Data Type    Description
unitsPerEm			USHORT      Granularity of the font's em square.
xMax				USHORT      Maximum X-coordinate for the entire font.
xMin				USHORT      Minimum X-coordinate for the entire font.
yMax				USHORT      Maximum Y-coordinate for the entire font.
yMin				USHORT      Minimum Y-coordinate for the entire font.
indexToLocFormat	SHORT       Used when processing the Index To Loc Table."
	| origin corner units indexToLocFormat |
	entry skip: 4. "Skip table version number"
	entry skip: 4. "Skip font revision number"
	entry skip: 4. "Skip check sum adjustment"
	entry skip: 4. "Skip magic number"
	entry skip: 2. "Skip flags"

	units := entry nextUShort.

	entry skip: 8. "Skip creation date"
	entry skip: 8. "Skip modification date"

	"Get min/max values of all glyphs"
	origin := entry nextShort @ entry nextShort.
	corner := entry nextShort @ entry nextShort.

	entry skip: 2. "Skip mac style"
	entry skip: 2. "Skip lowest rec PPEM"
	entry skip: 2. "Skip font direction hint"
	indexToLocFormat := entry nextShort.

	fontDescription setBounds: (origin corner: corner) unitsPerEm: units.
	^indexToLocFormat! !

!TTFontReader methodsFor: 'processing' stamp: 'ul 6/17/2011 12:31'!
processGlyphDataTable: entry offsets: offsetArray
	"Read the actual glyph data from the font.
	offsetArray contains the start offsets in the data for each glyph."
	| initialOffset |
	initialOffset := entry offset.
	glyphs := Array new: nGlyphs.
	1 to: nGlyphs do:[:i | 
		glyphs at: i put: (TTGlyph new glyphIndex: i-1)].
	'Reading glyph data' 
		displayProgressFrom: 1 
		to: nGlyphs 
		during:[:bar| | nContours glyphOffset origin corner glyphLength glyph nextOffset |

	1 to: nGlyphs do:[:glyphIndex |
		bar value: glyphIndex.
		glyph := glyphs at: glyphIndex.
		glyphOffset := offsetArray at: glyphIndex.
		nextOffset := offsetArray at: glyphIndex+1.
		glyphLength := nextOffset - glyphOffset.
		glyphLength = 0 ifFalse:[
			entry offset: initialOffset + glyphOffset.
			nContours := entry nextShort.
			origin := entry nextShort @ entry nextShort.
			corner := entry nextShort @ entry nextShort.
			glyph bounds: (origin corner: corner).
			nContours >= 0 ifTrue:[
				self processSimpleGlyph: glyph contours: nContours from: entry
			] ifFalse:[
				glyph := self processCompositeGlyph: glyph contours: nContours from: entry.
				glyphs at: glyphIndex put: glyph]]]
	].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalHeaderTable: entry
"
ascender           SHORT          Typographic ascent.
descender          SHORT          Typographic descent.
lineGap            SHORT          Typographic lineGap.
numberOfHMetrics   USHORT         Number hMetric entries in the HTMX
                                               Table; may be smaller than the total
                                             number of glyphs.
"
	| asc desc lGap numHMetrics |
	entry skip: 4. "Skip table version"
	asc := entry nextShort.
	desc := entry nextShort.
	lGap := entry nextShort.
	entry skip: 2. "Skip advanceWidthMax"
	entry skip: 2. "Skip minLeftSideBearing"
	entry skip: 2. "Skip minRightSideBearing"
	entry skip: 2. "Skip xMaxExtent"
	entry skip: 2. "Skip caretSlopeRise"
	entry skip: 2. "Skip caretSlopeRun"
	entry skip: 10. "Skip 5 reserved shorts"
	entry skip: 2. "Skip metricDataFormat"

	numHMetrics := entry nextUShort.

	fontDescription setAscender: asc descender: desc lineGap: lGap.
	^numHMetrics! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:40'!
processHorizontalMetricsTable: entry length: numHMetrics
	"Extract the advance width, left side bearing, and right
	side bearing for each glyph from the Horizontal Metrics Table."
	|  index lastAW glyph |
	index := 1.
	[index <= numHMetrics] whileTrue:[
		glyph := glyphs at: index.
		glyph advanceWidth: entry nextUShort.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index := index + 1].
	index = (nGlyphs +1) ifTrue:[^true].
	lastAW := (glyphs at: index-1) advanceWidth.

	[index <= nGlyphs] whileTrue:[
		glyph := glyphs at: index.
		glyph advanceWidth: lastAW.
		glyph leftSideBearing: entry nextShort.
		glyph updateRightSideBearing.
		index := index + 1].! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 11/2/1998 00:43'!
processIndexToLocationTable: entry format: indexToLocFormat
"glyphOffset    ULONG[numGlyphs]   An array that contains each glyph's
                                 offset into the Glyph Data Table.
"	| glyphOffset offset|
	glyphOffset := Array new: nGlyphs+1.
	1 to: nGlyphs+1 do:[:i|
		(indexToLocFormat = 0) ifTrue:[ "Format0: offset/2 is stored"
			offset := entry nextUShort * 2.
		] ifFalse:["Format1: store actual offset"
			offset := entry nextULong].
		glyphOffset at: i put: offset].
	^glyphOffset! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/19/2015 01:33'!
processKerningSubTableType0: entry

	| kp nKernPairs |
	nKernPairs := entry nextUShort.
	entry skip: 2. "Skip search range"
	entry skip: 2. "Skip entry selector"
	entry skip: 2. "Skip range shift"
	kernPairs := Array new: nKernPairs.
	1 to: nKernPairs do:[:i|
		kp := TTKernPair new.
		kp left: entry nextUShort.
		kp right: entry nextUShort.
		kp value: entry nextShort.
		kernPairs at: i put: kp].
	^true
! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/19/2015 01:34'!
processKerningTable: entry
	"Extract the kerning information for pairs of glyphs."
	| version  nTables covLow covHigh |
	version := entry nextUShort. "Skip table version"
	nTables := entry nextUShort. "Skip number of sub tables -- we're using the first one only"
	nTables = 0 ifTrue: [" This detection is hacky "
		entry skip: -4. "rewind"
		^ self processOSXKerningTable: entry].
	entry skip: 2. "Skip current subtable number"
	entry skip: 2. "Skip length of subtable"
	covHigh := entry nextByte.
	covLow := entry nextByte.

	"Make sure the format is right (kerning table and format type 0)"
	((covLow bitAnd: 2) = 2 or:[ covHigh ~= 0]) ifTrue:[^false].
	"Subtable"
	^ self processKerningSubTableType0: entry.
! !

!TTFontReader methodsFor: 'processing' stamp: ''!
processMaximumProfileTable: entry
"
numGlyphs         USHORT      The number of glyphs in the font.
"
	entry skip: 4. "Skip Table version number"
	nGlyphs := entry nextUShort.! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/7/2015 02:16'!
processNamingTable: entry
"copyright         CHARPTR     The font's copyright notice.
familyName        CHARPTR     The font's family name.
subfamilyName     CHARPTR     The font's subfamily name.
uniqueName        CHARPTR     A unique identifier for this font.
fullName          CHARPTR     The font's full name (a combination of
                                          familyName and subfamilyName).
versionName       CHARPTR     The font's version string.
"	| nRecords initialOffset format storageOffset  strings |
	strings := Array new: 8.
	strings atAllPut:''.
	initialOffset := entry offset.
	format := entry nextUShort.
	format = 0 ifFalse: [self error: 'Cannot handle format 1 naming tables'].
	"Get the number of name records"
	nRecords := entry nextUShort.
	"Offset from the beginning of this table"
	storageOffset := entry nextUShort + initialOffset.
	1 to: nRecords do:[:i| | pID sID lID nID length offset multiBytes string |
		pID := entry nextUShort.
		sID := entry nextUShort.
		lID := entry nextUShort.
		nID := entry nextUShort.
		length := entry nextUShort.
		offset := entry nextUShort.
		"Read only Macintosh or Microsoft strings"
		(pID = 1 or:[pID = 3 and:[sID = 1]]) ifTrue:[
			"MS uses Unicode all others single byte"
			multiBytes := pID = 3.
			string := entry stringAt: storageOffset + offset length: length multiByte: multiBytes.

			"Select only English names.
			Note: We prefer Macintosh strings about everything."
			((pID = 1 and: [lID = 0]) or: [pID = 3 and: [lID = 16r0409]]) ifTrue: [
				"Put the name at the right location."
				nID < strings size ifTrue:[
					(pID = 1 or:[(strings at: nID+1) = ''])
						ifTrue:[strings at: nID+1 put: string].
				].
			].
		].
	].
	fontDescription setStrings: strings.! !

!TTFontReader methodsFor: 'processing' stamp: 'ar 8/29/2009 19:19'!
processOS2Table: entry
"
	USHORT  	 version   	0x0004
	SHORT 	xAvgCharWidth 	 
	USHORT 	usWeightClass 	 
	USHORT 	usWidthClass 	 
	USHORT 	fsType 	 
	SHORT 	ySubscriptXSize 	 
	SHORT 	ySubscriptYSize 	 
	SHORT 	ySubscriptXOffset 	 
	SHORT 	ySubscriptYOffset 	 
	SHORT 	ySuperscriptXSize 	 
	SHORT 	ySuperscriptYSize 	 
	SHORT 	ySuperscriptXOffset 	 
	SHORT 	ySuperscriptYOffset 	 
	SHORT 	yStrikeoutSize 	 
	SHORT 	yStrikeoutPosition 	 
	SHORT 	sFamilyClass 	 
	BYTE 	panose[10] 	 
	ULONG 	ulUnicodeRange1 	Bits 0-31
	ULONG 	ulUnicodeRange2 	Bits 32-63
	ULONG 	ulUnicodeRange3 	Bits 64-95
	ULONG 	ulUnicodeRange4 	Bits 96-127
	CHAR 	achVendID[4] 	 
	USHORT 	fsSelection 	 
	USHORT 	usFirstCharIndex 	 
	USHORT 	usLastCharIndex 	 
	SHORT 	sTypoAscender 	 
	SHORT 	sTypoDescender 	 
	SHORT 	sTypoLineGap 	 
	USHORT 	usWinAscent 	 
	USHORT 	usWinDescent 	 
	ULONG 	ulCodePageRange1 	Bits 0-31
	ULONG 	ulCodePageRange2 	Bits 32-63
	SHORT 	sxHeight 	 
	SHORT 	sCapHeight 	 
	USHORT 	usDefaultChar 	 
	USHORT 	usBreakChar 	 
	USHORT 	usMaxContext 	 "
	| version fsSelection minAscii maxAscii asc desc lGap |
	version := entry nextShort. "table version"
	version = 0 ifTrue:[^self].
	entry skip: 60.
	fsSelection := entry nextUShort.
	minAscii := entry nextUShort.
	maxAscii := entry nextUShort.
	asc := entry nextShort.
	desc := entry nextShort.
	lGap := entry nextShort.
	fontDescription setTypographicAscender: asc descender: desc lineGap: lGap.! !

!TTFontReader methodsFor: 'processing' stamp: 'topa 3/19/2015 01:35'!
processOSXKerningTable: entry
	"From Apple's TrueType reference:
	
	Previous versions of the 'kern' table defined both the version and nTables fields in the header as UInt16 values and not UInt32 values. Use of the older format on OS X is discouraged (although AAT can sense an old kerning table and still make correct use of it). Microsoft Windows still uses the older format for the 'kern' table and will not recognize the newer one. Fonts targeted for OS X only should use the new format; fonts targeted for both OS X and Windows should use the old format.
	"
	"Extract the kerning information for pairs of glyphs."
	| version  nTables length coverage tupleIndex |
	version := entry nextULong. "Skip table version"
	nTables := entry nextULong. "Skip number of sub tables -- we're using the first one only"

	length := entry nextULong. "Skip length of subtable"
	coverage := entry nextUShort.
	tupleIndex := entry nextUShort. "Skip tuple index".
	"Make sure the format is right (kerning table and format type 0)"
	coverage = 0 ifFalse: [^false].
	
	"Subtable"
	^ self processKerningSubTableType0: entry.
! !

!TTFontReader methodsFor: 'processing' stamp: 'yo 3/19/2009 10:37'!
processSimpleGlyph: glyph contours: nContours from: entry

	| endPts  nPts iLength flags |
	endPts := Array new: nContours.
	1 to: nContours do:[:i| endPts at: i put: entry nextUShort].
	glyph initializeContours: nContours with: endPts.
	nContours = 0 ifTrue: [^ self].
	nPts := endPts last + 1.
	iLength := entry nextUShort. "instruction length"
	entry skip: iLength.
	flags := self getGlyphFlagsFrom: entry size: nPts.
	self readGlyphXCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	self readGlyphYCoords: entry glyph: glyph nContours: nContours flags: flags endPoints: endPts.
	glyph buildContours.! !

!TTFontReader methodsFor: 'public' stamp: 'sd 1/30/2004 15:24'!
readFrom: aStream

	| fontData headerEntry maxProfileEntry nameEntry indexLocEntry charMapEntry glyphEntry horzHeaderEntry horzMetricsEntry kerningEntry glyphOffset cmap numHMetrics indexToLocFormat |

	"Read the raw font byte data"
	aStream binary.
	fontData := aStream contents asByteArray.
	fontDescription := TTFontDescription new.

	"Search the tables required to build the font"
	(headerEntry := self getTableDirEntry: 'head' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a header table'].
	(maxProfileEntry := self getTableDirEntry: 'maxp' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a maximum profile table'].
	(nameEntry := self getTableDirEntry: 'name' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a name table'].
	(indexLocEntry := self getTableDirEntry: 'loca' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a relocation table'].
	(charMapEntry := self getTableDirEntry: 'cmap' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a character map table'].
	(glyphEntry := self getTableDirEntry: 'glyf' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a glyph table'].
	(horzHeaderEntry := self getTableDirEntry: 'hhea' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal header table'].
	(horzMetricsEntry := self getTableDirEntry: 'hmtx' from: fontData) == nil ifTrue:[
		^self error:'This font does not have a horizontal metrics table'].
	(kerningEntry := self getTableDirEntry: 'kern' from: fontData) == nil ifTrue:[
		Transcript cr; show:'This font does not have a kerning table';endEntry].


	"Process the data"
	indexToLocFormat := self processFontHeaderTable: headerEntry.
	self processMaximumProfileTable: maxProfileEntry.
	self processNamingTable: nameEntry.
	glyphOffset := self processIndexToLocationTable: indexLocEntry format: indexToLocFormat.
	cmap := self processCharacterMappingTable: charMapEntry.
	(cmap == nil or:[cmap value == nil])
		ifTrue:[^self error:'This font has no suitable character mappings'].
	self processGlyphDataTable: glyphEntry offsets: glyphOffset.
	numHMetrics := self processHorizontalHeaderTable: horzHeaderEntry.
	self processHorizontalMetricsTable: horzMetricsEntry length: numHMetrics.
	kerningEntry isNil 
		ifTrue:[kernPairs := #()]
		ifFalse:[self processKerningTable: kerningEntry].
	charMap := self processCharMap: cmap.
	fontDescription setGlyphs: glyphs mapping: charMap.
	fontDescription setKernPairs: kernPairs.
	^fontDescription! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:36'!
readGlyphXCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the x coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits xValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"If bit zero in the flag is set then this point is an on-curve
			point, if not, then it is an off-curve point."
			(flagBits bitAnd: 1) = 1 
				ifTrue:[ ttPoint type: #OnCurve]
				ifFalse:[ttPoint type: #OffCurve].
			"First we check to see if bit one is set.  This would indicate that
			the corresponding coordinate data in the table is 1 byte long.
			If the bit is not set, then the coordinate data is 2 bytes long."
			(flagBits bitAnd: 2) = 2 ifTrue:[ "one byte"
				xValue := entry nextByte.
				xValue := (flagBits bitAnd: 16)=16 ifTrue:[xValue] ifFalse:[xValue negated].
				ttPoint x: xValue.
			] ifFalse:[ "two byte"
				"If bit four is set, then this coordinate is the same as the
				last one, so the relative offset (of zero) is stored.  If bit
				is not set, then read in two bytes and store it as a signed value."
				(flagBits bitAnd: 16) = 16 ifTrue:[ ttPoint x: 0 ]
				ifFalse:[
					xValue := entry nextShort.
					ttPoint x: xValue]]].
		startPoint := endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/2/1998 01:37'!
readGlyphYCoords:entry glyph: glyph nContours: nContours flags: flags endPoints: endPts
	"Read the y coordinates for the given glyph from the font file."
	| startPoint endPoint flagBits yValue contour ttPoint |
	startPoint := 1.
	1 to: nContours do:[:i|
		contour := glyph contours at: i.
		"Get the end point"
		endPoint := (endPts at: i) + 1.
		"Store number of points"
		startPoint to: endPoint do:[:j|
			ttPoint := contour points at: (j - startPoint + 1).
			flagBits := flags at: j.
			"Check if this value one or two byte encoded"
			(flagBits bitAnd: 4) = 4 ifTrue:[ "one byte"
				yValue := entry nextByte.
				yValue := (flagBits bitAnd: 32)=32 ifTrue:[yValue] ifFalse:[yValue negated].
				ttPoint y: yValue.
			] ifFalse:[ "two byte"
				(flagBits bitAnd: 32) = 32 ifTrue:[ ttPoint y: 0 ]
				ifFalse:[
					yValue := entry nextShort.
					ttPoint y: yValue]]].
		startPoint := endPoint + 1]! !

!TTFontReader methodsFor: 'private' stamp: 'ar 11/1/1998 21:01'!
warn: aString
	Transcript cr; show: aString; endEntry.! !

!TTFontReader methodsFor: 'private' stamp: 'ar 4/10/2005 16:11'!
winToMac: index
	^ (index - 1) asCharacter squeakToMac asciiValue + 1! !

Object subclass: #TTFontTableDirEntry
	instanceVariableNames: 'tag fontData offset length checkSum'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTFontTableDirEntry commentStamp: '<historical>' prior: 0!
This class represents an entry in a truetype font table directory. Used by TTFontReader only.!

!TTFontTableDirEntry class methodsFor: 'instance creation' stamp: ''!
on: fontData at: index

	^self new on: fontData at: index! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextByte

	| value |
	value := fontData byteAt: offset.
	offset := offset + 1.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextBytes: numBytes into: array startingAt: byteOffset

	1 to: numBytes do:[:i|
		array at: i put: (fontData byteAt: byteOffset + i - 1)].! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextLong

	| value |
	value := fontData longAt: offset bigEndian: true.
	offset := offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextShort

	| value |
	value := fontData shortAt: offset bigEndian: true.
	offset := offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextULong

	| value |
	value := fontData unsignedLongAt: offset bigEndian: true.
	offset := offset + 4.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
nextUShort

	| value |
	value := fontData unsignedShortAt: offset bigEndian: true.
	offset := offset + 2.
	^value! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
offset
	^offset! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
offset: newOffset
	offset := newOffset! !

!TTFontTableDirEntry methodsFor: 'initialize-release' stamp: ''!
on: fd at: index

	fontData := fd.
	tag := fontData longAt: index bigEndian: true.
	checkSum := fontData longAt: index+4 bigEndian: true.
	offset := (fontData longAt: index+8 bigEndian: true) + 1.
	length := fontData longAt: index+12 bigEndian: true.! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: ''!
skip: n
	"Skip n bytes"
	offset := offset + n.! !

!TTFontTableDirEntry methodsFor: 'accessing' stamp: 'ar 11/1/1998 23:37'!
stringAt: stringOffset length: byteLength multiByte: aBoolean

	| string index stringLength |
	aBoolean ifFalse:[
		stringLength := byteLength.
		string := String new: stringLength.
		index := stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + i - 1))].
		^string
	] ifTrue:[
		stringLength := byteLength // 2.
		string := String new: stringLength.
		index := stringOffset.
		1 to: stringLength do:[:i|
			string at: i put: (Character value: (fontData byteAt: index + 1)).
			index := index + 2].
		^string]! !

Object subclass: #TTGlyph
	instanceVariableNames: 'bounds contours advanceWidth leftSideBearing rightSideBearing glyphIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTGlyph commentStamp: '<historical>' prior: 0!
This class represents a glyph of a TrueType font.

Instance variables:
	bounds			<Rectangle>	The receiver's bounds
	contours		<Array of: PointArray> The compressed contours in the receiver
	advanceWidth	<Integer>	advance width of the glyph
	leftSideBearing	<Integer>	left side bearing
	rightSideBearing <Integer>	right side bearing
	glyphIndex 		<Integer>	the original index of the glyph (used for kerning)!

TTGlyph subclass: #TTCompositeGlyph
	instanceVariableNames: 'glyphs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCompositeGlyph commentStamp: '<historical>' prior: 0!
This class represents a composite TrueType glyph, e.g.one which contains many simple TTGlyphs.!

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:21'!
addGlyph: aGlyph transformation: aMatrix
	glyphs := glyphs copyWith: (aMatrix -> aGlyph)! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'yo 12/10/2007 17:15'!
basicGlyphs

	^ glyphs
! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'topa 3/10/2015 21:01'!
basicGlyphs: colOfAssocs

	glyphs := colOfAssocs
! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 7/30/2009 17:06'!
buildAllContours
	"Build the contours in all non-composite glyphs."
	glyphs do:[:assoc| assoc value buildAllContours].! !

!TTCompositeGlyph methodsFor: 'private' stamp: 'ar 11/2/1998 01:43'!
computeContours
	| out |
	out := WriteStream on: (Array new: glyphs size * 4).
	self glyphsAndTransformationsDo:[:glyph :transform|
		glyph contours do:[:ptArray|
			out nextPut: (transform localPointsToGlobal: ptArray).
		].
	].
	^out contents! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:43'!
contours
	^contours ifNil:[contours := self computeContours]! !

!TTCompositeGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:27'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours := nil.! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:20'!
glyphs

	^glyphs collect:[:assoc| assoc value].! !

!TTCompositeGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:28'!
glyphsAndTransformationsDo: aBlock
	glyphs do:[:assoc|
		aBlock value: assoc value value: assoc key.
	].! !

!TTCompositeGlyph methodsFor: 'initialize' stamp: 'ar 11/2/1998 01:20'!
initialize
	glyphs := #().! !

!TTCompositeGlyph methodsFor: 'testing' stamp: ''!
isComposite
	^true! !

!TTCompositeGlyph methodsFor: 'initialize' stamp: 'nice 12/27/2009 03:10'!
referenceVertexAt: index
	"Only used while reading before constructing contours"
	| i |
	i := index.
	self glyphsAndTransformationsDo: [:glyph :transform | | p |
		p := glyph referenceVertexAt: i.
		p isPoint
			ifTrue: [^transform localPointToGlobal: p].
		i := i - p].
	self error: ['this should not happen']! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
advanceWidth
	^advanceWidth! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
advanceWidth: aNumber
	advanceWidth := aNumber.! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:29'!
asFormWithScale: scale ascender: ascender descender: descender
	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: Color black
		bgColor: Color white
		depth: 8
		replaceColor: true.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 6/23/2003 18:27'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: false.
! !

!TTGlyph methodsFor: 'converting' stamp: 'yo 5/7/2004 10:37'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag

	^ self
		asFormWithScale: scale
		ascender: ascender
		descender: descender
		fgColor: fgColor
		bgColor: bgColor
		depth: depth
		replaceColor: replaceColorFlag
		lineGlyph: nil
		lingGlyphWidth: 0
		emphasis: 0.! !

!TTGlyph methodsFor: 'converting' stamp: 'mt 8/16/2016 18:13'!
asFormWithScale: scale ascender: ascender descender: descender fgColor: fgColor bgColor: bgColor depth: depth replaceColor: replaceColorFlag lineGlyph: lineGlyph lingGlyphWidth: lWidth emphasis: code

	| form canvas newScale |
	form := Form extent: (advanceWidth @ (ascender - descender) * scale) rounded depth: depth.
	form fillColor: bgColor.
	canvas := form getCanvas asBalloonCanvas.
	canvas aaLevel: 4.
	canvas transformBy: (MatrixTransform2x3 withScale: scale asPoint * (1 @ -1)).
	canvas transformBy: (MatrixTransform2x3 withOffset: 0 @ ascender negated).
	canvas
		drawGeneralBezierShape: self contours
		color: fgColor 
		borderWidth: 0 
		borderColor: fgColor.
	((code bitAnd: 4) ~= 0 or: [(code bitAnd: 16) ~= 0]) ifTrue: [
		newScale := (form width + 1) asFloat / lineGlyph calculateWidth asFloat.
		canvas transformBy: (MatrixTransform2x3 withScale: (newScale / scale)@1.0).

		(code bitAnd: 4) ~= 0 ifTrue: [
			canvas
				drawGeneralBezierShape: lineGlyph contours
				color: fgColor 
				borderWidth: 0 
				borderColor: fgColor.
		].

		(code bitAnd: 16) ~= 0 ifTrue: [
			canvas transformBy: (MatrixTransform2x3 withOffset: 0@(ascender // 2)).
			canvas
				drawGeneralBezierShape: lineGlyph contours
				color: fgColor 
				borderWidth: 0 
				borderColor: fgColor.
		].
	].

	replaceColorFlag ifTrue: [
		form replaceColor: bgColor withColor: Color transparent.
	].
	^ form! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds
	^bounds! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/1/1998 22:25'!
bounds: aRectangle
	bounds := aRectangle! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 7/30/2009 17:06'!
buildAllContours
	"Build the contours in all non-composite glyphs."
	^self buildContours! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'nice 12/27/2009 03:10'!
buildContours
	"Build the contours in the receiver glyph.
	The contour is constructed by converting the points
	form each contour into an absolute value and then
	compressing the contours into PointArrays."
	| tx ty |
	tx := ty := 0.
	contours := contours collect:[:contour| | points |
		contour isCollection ifTrue:[^self]. "already built"
		points := contour points.
		points do:[:pt|
			pt x: (tx := tx + pt x).
			pt y: (ty := ty + pt y)].
		contour asCompressedPoints].! !

!TTGlyph methodsFor: 'private' stamp: 'yo 5/7/2004 10:38'!
calculateWidth

	| min max |
	min := SmallInteger maxVal.
	max := SmallInteger minVal.
	self contours do: [:a | a do: [:p |
		p x > max ifTrue: [max := p x].
		p x < min ifTrue: [min := p x].
	]].
	^ max - min.
! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
contours
	^contours! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
contours: aCollection
	contours := aCollection asArray.! !

!TTGlyph methodsFor: 'private' stamp: 'ar 5/25/2000 18:01'!
display
	| canvas |
	canvas := Display getCanvas.
	self contours do:[:ptArray|
		1 to: ptArray size by: 3 do:[:i|
			canvas line: (ptArray at: i) // 10
					to: (ptArray at: i+2) // 10
					width: 1 color: Color black.
		].
	].! !

!TTGlyph methodsFor: 'private' stamp: 'ar 11/14/1998 20:22'!
flipAroundY
	bounds := (bounds origin x @ bounds corner y negated) corner:
				(bounds corner x @ bounds origin y negated).
	contours := self contours collect:[:contour| contour collect:[:pt| pt x @ pt y negated]].! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
glyphIndex
	^glyphIndex! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
glyphIndex: anInteger
	glyphIndex := anInteger! !

!TTGlyph methodsFor: 'accessing' stamp: 'ar 11/2/1998 01:26'!
glyphsAndTransformationsDo: aBlock
	aBlock value: self value: MatrixTransform2x3 identity! !

!TTGlyph methodsFor: 'initialize-release' stamp: 'ar 11/1/1998 22:25'!
initialize

	bounds := 0@0 corner: 0@0.
	contours := #().
	advanceWidth := 0.
	leftSideBearing := 0.
	rightSideBearing := 0.! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:42'!
initializeContours: numContours with: endPoints
	"Initialize the contours for creation of the glyph."
	| startPt pts endPt |
	contours := Array new: numContours.
	startPt := -1.
	1 to: numContours do:[:i|
		endPt := endPoints at: i.
		pts := Array new: endPt - startPt.
		1 to: pts size do:[:j| pts at: j put: TTPoint new].
		contours at: i put: (TTContourConstruction on: pts).
		startPt := endPt].! !

!TTGlyph methodsFor: 'testing' stamp: ''!
isComposite
	^false! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
leftSideBearing
	^leftSideBearing! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
leftSideBearing: aNumber
	leftSideBearing := aNumber.! !

!TTGlyph methodsFor: 'printing' stamp: 'tk 9/13/1999 09:54'!
printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPut:$(;
		print: (contours ifNil: [0] ifNotNil: [contours size]);
		nextPut:$).! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'nice 12/27/2009 03:10'!
referenceVertexAt: index
	"Only used while reading before constructing contours"
	| count |
	count := 0.
	contours do: [:construction | | vertices |
		vertices := construction points.
		index - count > vertices size
			ifTrue: [count := count + vertices size]
			ifFalse: [^(vertices at: index - count) asPoint]].
	^count! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
rightSideBearing
	^rightSideBearing! !

!TTGlyph methodsFor: 'accessing' stamp: ''!
rightSideBearing: aNumber
	rightSideBearing := aNumber.! !

!TTGlyph methodsFor: 'private-initialization' stamp: 'ar 11/1/1998 22:27'!
updateRightSideBearing
	"Update the right side bearing value"
	"@@: Is the following really correct?!!?!!"
	rightSideBearing := advanceWidth - leftSideBearing - bounds corner x + bounds origin x! !

Object subclass: #TTKernPair
	instanceVariableNames: 'left right value mask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTKernPair commentStamp: '<historical>' prior: 0!
A TTKernPair represents a TrueType kerning pair.

Instance variables:
	left	<Integer>	The glyph index for the left character.
	right <Integer>	The glyph index for the right character.
	value <Integer>	The amount of kerning.
	mask <Integer>	An efficient representation for the left and the right value.!

!TTKernPair class methodsFor: 'accessing' stamp: ''!
maskFor: left with: right
	^(left bitShift: 12) + right! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
left
	^left! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
left: aNumber

	left := aNumber! !

!TTKernPair methodsFor: 'accessing' stamp: 'ar 11/1/1998 20:08'!
mask
	^mask ifNil:[mask := self class maskFor: left with: right]! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
right
	^right! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
right: aNumber

	right := aNumber! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
value
	^value! !

!TTKernPair methodsFor: 'accessing' stamp: ''!
value: aNumber

	value := aNumber! !

Object subclass: #TTPoint
	instanceVariableNames: 'x y type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Support'!

!TTPoint commentStamp: '<historical>' prior: 0!
A representation of a TrueType point which includes a 'type' flag defining whether this point is an 'on' or an 'off' curve point.!

!TTPoint methodsFor: 'converting' stamp: ''!
asPoint
	^x@y! !

!TTPoint methodsFor: 'printing' stamp: ''!
printOn: aStream

	aStream 
		nextPutAll: self class name;
		nextPut:$(;
		print: x;
		nextPut:$@;
		print: y;
		nextPut:$|;
		print: type;
		nextPut:$)! !

!TTPoint methodsFor: 'accessing' stamp: ''!
type
	^type! !

!TTPoint methodsFor: 'accessing' stamp: ''!
type: aSymbol

	type := aSymbol! !

!TTPoint methodsFor: 'accessing' stamp: ''!
x
	^x! !

!TTPoint methodsFor: 'accessing' stamp: ''!
x: aNumber

	x := aNumber! !

!TTPoint methodsFor: 'accessing' stamp: ''!
y
	^y! !

!TTPoint methodsFor: 'accessing' stamp: ''!
y: aNumber
	y := aNumber! !

!HostFont class methodsFor: '*TrueType' stamp: 'yo 3/17/2004 00:39'!
initForSubtitles
"
	HostFont initForSubtitles
"

	HostFont textStyleFrom: 'Verdana' sizes: #(18 20 22 24 26 28) ranges: HostFont defaultRanges.

	StrikeFontSet installExternalFontFileName: 'greekFont.out' encoding: GreekEnvironment leadingChar encodingName: #Greek textStyleName: #DefaultMultiStyle.


	TTCFontReader encodingTag: SimplifiedChineseEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\simhei.TTF'.

	TTCFontReader encodingTag: JapaneseEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\msgothic.TTC'.

	TTCFontReader encodingTag: KoreanEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\gulim.TTC'.
! !

!HostFont class methodsFor: '*TrueType' stamp: 'yo 12/2/2004 12:50'!
unloadAsianTT
"
	self unloadAsianTT
"

	TTCFontSet removeStyleName: 'MultiSimHei'.
	TTCFontSet removeStyleName: 'MultiMSGothic'.
	TTCFontSet removeStyleName: 'MultiGulim'.
! !

AbstractFont subclass: #TTCFont
	instanceVariableNames: 'ttcDescription pointSize foregroundColor cache derivatives fallbackFont height ascent descent maxAscii colorToCacheMap'
	classVariableNames: 'GlyphCacheData GlyphCacheIndex GlyphCacheReady GlyphCacheSize NamesToIndexes Registry Scale ShutdownList'
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCFont commentStamp: 'nk 4/2/2004 11:32' prior: 0!
I represent a font that uses TrueType derived glyph.  Upon a request for glyph for a character through a call to #formOf: (or #widthOf:), I first search corresponding glyph in the cache.  If there is not, it creates a 32bit depth form with the glyph.

  The cache is weakly held.  The entries are zapped at full GC.

Structure:
 ttcDescription	TTFontDescription -- The Squeak data structure for a TrueType font data file.
 pointSize		Number -- Nominal Em size in points. Conversion to pixel sizes depends on the definition of TextStyle class>>pixelsPerInch.
 foregroundColor	Color -- So far, this font need to know the glyph color in cache.
 cache			WeakArray of <Color -> <Array(256) of glyph>>
 derivatives		Array -- stores the fonts in the same family but different emphasis.
!

TTCFont subclass: #LinedTTCFont
	instanceVariableNames: 'emphasis lineGlyph contourWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!LinedTTCFont class methodsFor: 'as yet unclassified' stamp: 'ar 11/14/2006 15:18'!
fromTTCFont: aTTCFont emphasis: code

	| inst |
	inst := self new.
	inst ttcDescription: aTTCFont ttcDescription.
	inst pointSize: aTTCFont pointSize.
	inst emphasis: (aTTCFont emphasis bitOr: code).
	inst lineGlyph: (aTTCFont ttcDescription at: $_).

	^ inst.
! !

!LinedTTCFont methodsFor: 'private' stamp: 'yo 5/7/2004 12:49'!
computeForm: char

	| ttGlyph scale |

	char = Character tab ifTrue: [^ super computeForm: char].

	"char = $U ifTrue: [self doOnlyOnce: [self halt]]."
	scale := self pixelSize asFloat / (ttcDescription ascender - ttcDescription descender).
	ttGlyph := ttcDescription at: char.
	^ ttGlyph asFormWithScale: scale ascender: ttcDescription ascender descender: ttcDescription descender fgColor: foregroundColor bgColor: Color transparent depth: self depth replaceColor: false lineGlyph: lineGlyph lingGlyphWidth: contourWidth emphasis: emphasis! !

!LinedTTCFont methodsFor: 'accessing' stamp: 'yo 5/6/2004 19:56'!
emphasis

	^ emphasis.
! !

!LinedTTCFont methodsFor: 'accessing' stamp: 'yo 5/6/2004 19:18'!
emphasis: code

	emphasis := code.
! !

!LinedTTCFont methodsFor: 'initialize' stamp: 'yo 5/7/2004 11:26'!
lineGlyph: aGlyph

	lineGlyph := aGlyph.
	contourWidth := aGlyph calculateWidth.
! !

TTCFont subclass: #MultiTTCFont
	instanceVariableNames: 'map cacheIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!MultiTTCFont class methodsFor: 'as yet unclassified' stamp: 'yo 12/10/2002 18:34'!
cacheAllNil
"
	self cacheAllNil
"
	self allInstances do: [:inst |
		inst cache do: [:e |
			e third ifNotNil: [^ false].
		].
	].

	^ true.
! !

!MultiTTCFont methodsFor: 'as yet unclassified' stamp: 'yo 12/10/2002 18:08'!
access: char at: index

	| wcache entry |
	wcache := self cache.
	entry := wcache at: index.
	wcache replaceFrom: index to: wcache size - 1 with: wcache startingAt: index + 1.
	wcache at: wcache size put: entry.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
at: char put: form

	| ind triplet |
	triplet := Array with: char asciiValue with: foregroundColor with: form.
	GlyphCacheData at: (GlyphCacheIndex := GlyphCacheIndex \\ GlyphCacheSize + 1) put: triplet.

	ind := self indexFor: char.
	map at: char asciiValue put: ind.
	self cache at: ind put: triplet.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/19/2015 02:48'!
cache
	(cache isNil or: [cache size ~= 512]) ifTrue: [self recreateCache]. "old weak-array caching"
	^cache! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
flushCache

	super flushCache.
	map := IdentityDictionary new: 512.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'xinwang 10/23/2010 18:30'!
foregroundColor: fgColor
	"Install the given foreground color"
	foregroundColor := fgColor.! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
formOf: char

	| newForm |
	cache ifNil: [ self recreateCache ].
	foregroundColor ifNil: [ self foregroundColor: Color black ].

	self hasCached: char ifTrue: [:form |
		^ form.
	].

	newForm := self computeForm: char.
	self at: char put: newForm.
	^ newForm.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
glyphInfoOf: char into: glyphInfoArray

	| newForm |
	self hasCached: char ifTrue: [:form |
		glyphInfoArray at: 1 put: form;
			at: 2 put: 0;
			at: 3 put: form width;
			at: 4 put: (self ascentOf: char);
			at: 5 put: self.
		^ glyphInfoArray.
	].

	newForm := self computeForm: char.
	self at: char put: newForm.

	glyphInfoArray at: 1 put: newForm;
		at: 2 put: 0;
		at: 3 put: newForm width;
		at: 4 put: (self ascentOf: char);
		at: 5 put: self.
	^ glyphInfoArray.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
hasCached: char ifTrue: aBlock

	| value triplet |
	value := char asciiValue.
	triplet := cache at: (map at: value ifAbsent: [^ false]).
	triplet ifNil: [^ false].
	(triplet at: 1) ~= value ifTrue: [^ false].
	(triplet at: 2) ~= foregroundColor ifTrue: [^ false].
	^ aBlock value: (triplet at: 3).
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
indexFor: char

	| triplet |
	map size > 511 ifTrue: [
		cacheIndex := 512 atRandom.
		triplet := self cache at: cacheIndex.
		triplet ifNotNil: [map removeKey: (triplet at: 1) ifAbsent: []].
		^ cacheIndex
	].
	^ (cacheIndex := cacheIndex + 1 \\ 512) + 1.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:48'!
initialize

	super initialize.
	cacheIndex := 511.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'topa 3/10/2015 01:49'!
recreateCache

	cache := WeakArray new: 512.
	map := IdentityDictionary new: 512.
! !

!MultiTTCFont methodsFor: 'all' stamp: 'yo 4/23/2007 13:30'!
widthOf: char

	^ (self formOf: char) width.
! !

!TTCFont class methodsFor: 'objects from disk' stamp: 'nk 6/25/2003 13:42'!
classVersion
	"Version 0 had pixelSize; version 1 changed it to pointSize"
	^1! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 8/1/2005 15:06'!
familyName: n pointSize: s emphasis: code

	"(TTCFont familyName: 'BitstreamVeraSans' pointSize: 12 emphasis: 0)"
	| t ret index |
	t := self registry at: n asSymbol ifAbsent: [#()].
	t isEmpty ifTrue: [
		t := (TextConstants at: #DefaultTextStyle) fontArray.
		ret := t first.
		ret pointSize >= s ifTrue: [^ ret emphasis: code].
		index := 2.
		[index <= t size and: [(t at: index) pointSize <= s]] whileTrue: [
			ret := t at: index.
			index := index + 1.
		].
		^ ret emphasis: code.
	].
	^ ((TextStyle named: n) addNewFontSize: s) emphasis: code.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 6/23/2003 18:44'!
family: f size: s

	^ self allInstances detect: [:a | a familyName = f and: [a pointSize = s]] ifNone: [nil].
! !

!TTCFont class methodsFor: 'file list services' stamp: 'nk 7/16/2003 15:40'!
fileReaderServicesForFile: fullName suffix: suffix 
	^(suffix = 'ttf')  | (suffix = '*') 
		ifTrue: [ self services ]
		ifFalse: [ #() ]! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 12/13/2002 12:20'!
getExistings: fontArray

	| result em |
	result := OrderedCollection new.
	result add: fontArray.
	1 to: 3 do: [:i |
		em := (fontArray collect: [:f | f emphasized: i]).
		(em at: 1) ~= (fontArray at: 1) ifTrue: [
			result add: em.
		].
	].
	^ result asArray.
! !

!TTCFont class methodsFor: 'other' stamp: 'topa 3/19/2015 03:25'!
indexOfSubfamilyName: aName
	| decoded |

	"decodeStyleName will consume all the modifiers and leave nothing if everything was recognized."
	decoded := TextStyle decodeStyleName: aName.
	decoded second isEmpty ifTrue: [ ^decoded first ].

	"If you get a halt here - please add the missing synonym to the lookup table in TextStyle>>decodeStyleName: ."
	
	self error: 'please add the missing synonym ', aName, ' to the lookup table in TextStyle>>decodeStyleName:'.

	^0.! !

!TTCFont class methodsFor: 'class initialization' stamp: 'ar 9/5/2010 09:18'!
initialize
"
	self initialize
"

	| tt |
	self allSubInstancesDo:[:fnt| fnt flushCache].
	GlyphCacheSize := 512.
	GlyphCacheData := Array new: GlyphCacheSize.
	GlyphCacheIndex := 0.
	GlyphCacheReady := true.
	
	tt := TTFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].
	FileServices registerFileReader: self.

	Smalltalk addToShutDownList: self.! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:46'!
isCacheAllNil
"
	self cacheAllNil
"
	self allInstances do: [:inst |
		inst cache do: [:e |
			e ifNotNil: [^ false].
		].
	].

	^ true.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'yo 11/8/2004 19:27'!
newTextStyleFromTTFile: fileName
	"Create a new TextStyle from specified file name.  On certain versions of Windows, you can evaluate following to get Arial font into the image.  On other platforms, wait and see someone implements the support code for FontPlugin then we can start relying on the generic font lookup mechanism.
	TTCFontReader encodingTag: 0.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\symbol.TTF'.
	"

	| description |
	description := TTFontDescription addFromTTFile: fileName.
	^ self newTextStyleFromTT: description.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'tb 6/24/2003 17:12'!
newTextStyleFromTTStream: readStream
"
"

	| description |
	description := TTFontDescription addFromTTStream: readStream.
	^ self newTextStyleFromTT: description.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'nice 12/27/2009 03:11'!
newTextStyleFromTT: description 
	"Create a new TextStyle from specified TTFontDescription instance."

	| array |
	array := self pointSizes collect: 
					[:pt | | f | 
					f := self new.
					f ttcDescription: description.
					f pointSize: pt].
	^self reorganizeForNewFontArray: array name: array first name asSymbol! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 20:18'!
pointSizes

	"The default sizes that are created when a TextStyle is created.  You can add new sizes by the new-size feature."
	^ #(9 12 15 24 36).
! !

!TTCFont class methodsFor: 'other' stamp: 'ar 11/14/2006 15:19'!
recreateCache
"
	self recreateCache.
"
	self allSubInstances do: [:inst | inst flushCache].
	Smalltalk garbageCollect.
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 11:11'!
registerAll
"
	TTCFont registerAll
"

	TextStyle allInstancesDo: [:e |
		(e fontArray first isMemberOf: TTCFont) ifTrue: [
			self register: e fontArray at: e fontArray first familyName asSymbol.
		].
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 10:52'!
register: anObject at: symbolName

	self registry at: symbolName put: anObject.
! !

!TTCFont class methodsFor: 'other' stamp: 'nice 8/27/2010 21:07'!
registry
	^ Registry
		ifNil: [Registry := IdentityDictionary new]
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 5/7/2004 08:09'!
removeAllDerivatives
"
	self removeAllDerivatives
"

	self allInstances do: [:s |
		s textStyle ifNotNil: [
			s textStyle fontArray do: [:f |
				f derivativeFont: nil at: 0.
			].
		].
	].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 11/30/2002 22:37'!
removeStyleName: aString

	TextConstants removeKey: aString asSymbol ifAbsent: [].
	TTFontDescription removeDescriptionNamed: aString asString.
! !

!TTCFont class methodsFor: 'instance creation' stamp: 'topa 3/19/2015 03:01'!
reorganizeForNewFontArray: array name: styleName

	| style existings regular altName |
	(TextConstants includesKey: styleName) ifFalse: [
		TextConstants at: styleName put: (TextStyle fontArray: array).
		^ TextConstants at: styleName.
	].
 
	"There is a text style with the name I want to use.  See if it is a TTC font..."
	style := TextConstants at: styleName.
	style isTTCStyle ifFalse: [
		altName := ((array at: 1) name, 'TT') asSymbol.
		^ self reorganizeForNewFontArray: array name: altName.
	].

	existings := (self getExistings: style fontArray), (Array with: array).
	regular := existings detect: [:e | (e at: 1) isRegular] ifNone: [existings at: 1].

	regular do: [:r |
		r addLined: r.
	].

	"The existing array may be different in size than the new one."
	existings do: [:e |
		(e at: 1) isRegular ifFalse: [
			regular do: [ :r | | f |
				f := e detect: [ :ea | ea pointSize = r pointSize ] ifNone: [ ].
				f ifNotNil: [ r derivativeFont: f ].
			].
		].
	].

	style newFontArray: regular.
	self register: regular at: styleName.
	self recreateCache.	
	^ style.
! !

!TTCFont class methodsFor: 'other' stamp: 'nk 7/18/2004 15:35'!
repairDerivativeFonts
	"Fix the cases where the derivatives are a different size than the originals."

	"
	TTCFont repairDerivativeFonts.
	"
	self allInstancesDo: [ :font | font pointSize: font pointSize ].
	Preferences refreshFontSettings.! !

!TTCFont class methodsFor: 'file list services' stamp: ''!
serviceInstallTrueTypeFontStyle
	"Return a service to install a true type font as a text style"

	^ SimpleServiceEntry
		provider: self
		label: 'install ttf style'
		selector: #newTextStyleFromTTFile: 
		description: 'install a true type font as a text style'
		buttonLabel: 'install ttf'! !

!TTCFont class methodsFor: 'file list services' stamp: ''!
services
	"Return a set of services for use in FileList"

	^ Array with: self serviceInstallTrueTypeFontStyle! !

!TTCFont class methodsFor: 'class initialization' stamp: 'ar 11/14/2006 15:28'!
shutDown
	"Flush the glyph cache"
	GlyphCacheData atAllPut: nil.
	GlyphCacheIndex := 0.
	ShutdownList ifNotNil:[ShutdownList do:[:fnt| fnt flushCache]].
	ShutdownList := WeakSet new.
! !

!TTCFont class methodsFor: 'class initialization' stamp: 'ar 9/5/2010 09:18'!
unload

	FileServices unregisterFileReader: self! !

!TTCFont class methodsFor: 'other' stamp: 'yo 3/17/2005 10:53'!
unregister: symbolName

	self registry removeKey: symbolName ifAbsent: [].
! !

!TTCFont class methodsFor: 'other' stamp: 'yo 6/23/2003 19:50'!
version

	^ '6.0'.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 5/7/2004 12:30'!
addLined

	self addLined: self.
	self derivativeFonts do: [:e |
		e ifNotNil: [self addLined: e].
	].
! !

!TTCFont methodsFor: 'private' stamp: 'yo 5/7/2004 12:23'!
addLined: aTTCFont

	| l |
	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 4.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 16.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 20.
	self derivativeFont: l at: l emphasis.
! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:25'!
ascent
	^ascent ifNil:[ascent := (ttcDescription typographicAscender * self pixelScale) truncated].! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 7/31/2004 19:30'!
ascentOf: aCharacter

"	(self hasGlyphFor: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont ascentOf: aCharacter.
		].
	].
"
	^ self ascent.
! !

!TTCFont methodsFor: 'private' stamp: 'nice 3/20/2010 09:16'!
at: char put: form
	| assoc |
	assoc := foregroundColor -> form.
	GlyphCacheData at: (GlyphCacheIndex := GlyphCacheIndex \\ GlyphCacheSize + 1) put: assoc.
	cache at: (char asInteger + 1) put: assoc.
	^form! !

!TTCFont methodsFor: 'private' stamp: 'topa 6/1/2016 11:29'!
basicHasGlyphOf: aCharacter
	"Answer whether this font includes a glyph for the given character"
	^ aCharacter charCode <= self maxAscii
		and:[
			aCharacter = Character null
				ifTrue: [(self formOf: aCharacter) notNil]
				ifFalse: [
					"this should also be
						 (self formOf: aCharacter) notNil
					but it does not work, the fallback glyph -- if present -- is always found instead.
					So we fake. if aCharacter is the same form as Character null aka 0, we assume absence."
					(self characterFormAt: aCharacter) bits ~= self fallbackForm bits]]! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:19'!
cache
	^cache! !

!TTCFont methodsFor: 'character shapes' stamp: 'nk 11/3/2004 10:02'!
characterFormAt: character 
	"Answer a Form copied out of the glyphs for the argument,  
	character. Use a cached copy if possible."

	^self formOf: character! !

!TTCFont methodsFor: 'html' stamp: 'dgd 9/2/2005 12:39'!
closeHtmlOn: aStream 
	"put on the given stream the tag to close the html  
	representation of the receiver"
	self htmlSize isZero
		ifFalse: [aStream nextPutAll: '</font>']! !

!TTCFont methodsFor: 'private' stamp: 'ar 8/29/2009 19:16'!
computeForm: char
	"Compute the glyph form for the given character"
	^ttcDescription renderGlyph: char height: self height fgColor: foregroundColor bgColor: Color transparent depth: self depth! !

!TTCFont methodsFor: 'objects from disk' stamp: 'nk 4/2/2004 11:29'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	"If we're reading in an old version with a pixelSize instance variable, convert it to a point size."

	"Deal with the change from pixelSize to pointSize, assuming the current monitor dpi."
	varDict at: 'pixelSize' ifPresent: [ :x | 
		pointSize := (TextStyle pixelsToPoints: x) rounded.
	].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
copy

	^ self.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:35'!
deepCopy

	^ self.
! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:35'!
depth

	^ 32.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 5/6/2004 19:54'!
derivativeFontArray

	^ derivatives.
! !

!TTCFont methodsFor: 'friend' stamp: 'nk 9/1/2004 13:01'!
derivativeFonts

	derivatives ifNil: [^ #()].
	^derivatives copyWithout: nil! !

!TTCFont methodsFor: 'friend' stamp: 'yo 8/3/2005 07:46'!
derivativeFont: aTTCFont

	| index |
	index := self indexOfSubfamilyName: (aTTCFont subfamilyName).
	index < 1 ifTrue: [
		^ self "inform: 'unknown sub family name.  This font will be skipped'".
	].

	self derivativeFont: aTTCFont at: index.

	self addLined: aTTCFont.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 5/7/2004 12:40'!
derivativeFont: aTTCFont at: index

	| newDeriv |
	aTTCFont ifNil: [derivatives := nil. ^ self].
	derivatives ifNil: [derivatives := Array new: 32].
	derivatives size < 32 ifTrue: [
		newDeriv := Array new: 32.
		newDeriv replaceFrom: 1 to: derivatives size with: derivatives.
		derivatives := newDeriv.
	].
	derivatives at: index put: aTTCFont.
! !

!TTCFont methodsFor: 'initialize' stamp: 'tween 4/23/2006 22:52'!
derivativeFont: aNewlyCreatedDerivativeFont mainFont: aMainFont

	self derivativeFont: aNewlyCreatedDerivativeFont at: aMainFont emphasis.
	aNewlyCreatedDerivativeFont emphasis: aMainFont emphasis.
	aNewlyCreatedDerivativeFont lineGlyph: (aMainFont ttcDescription at: $_).! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/29/2009 18:26'!
descent
	"One is added to make sure the gap between lines is filled.  If we don't add, multi line selection in a text pane look ugly."
	^descent ifNil:[descent := self height - self ascent].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
descentKern

	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 7/31/2004 19:30'!
descentOf: aCharacter

"	(self hasGlyphFor: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont descentOf: aCharacter.
		].
	]."
	^ self descent.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 1/7/2005 12:04'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta 

	^ self displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.
! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:40'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

	| form glyphInfo destX destY hereX nextX actualFont |
	destX := aPoint x.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do: [:charIndex |
		self glyphInfoOf: (aString at: charIndex) into: glyphInfo.
		form := glyphInfo at: 1.
		hereX := glyphInfo at: 2.
		nextX := glyphInfo at: 3.
		(actualFont := glyphInfo at: 5) ==  aBitBlt lastFont
			ifFalse: [actualFont installOn: aBitBlt].
		destY := baselineY - (glyphInfo at: 4). 
		aBitBlt sourceForm: form.
		aBitBlt destX: destX.
		aBitBlt destY: destY.
		aBitBlt sourceX: hereX; sourceY: 0.
		aBitBlt width: nextX - hereX.
		aBitBlt height: form height.
		aBitBlt copyBits.
		destX := destX + (nextX - hereX) + kernDelta.
	].
	^ destX @ destY
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 5/6/2004 19:25'!
emphasis
	"Answer the emphasis code (0 to 3) corresponding to my subfamily name"
	^self indexOfSubfamilyName: self subfamilyName

! !

!TTCFont methodsFor: 'accessing' stamp: 'nice 8/27/2010 21:06'!
emphasis: code

	code > 3 ifTrue: [^ self].
	code = 0 ifTrue: [^ self].
	derivatives ifNil: [^ self].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 8/1/2005 12:14'!
emphasized: code

	code = 0 ifTrue: [^ self].
	derivatives ifNil: [^ self].
	(((code bitAnd: 20) ~= 0) and: [
		derivatives size < code or: [(derivatives at: code) isNil]]) ifTrue: [
		self addLined.
	].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/15/2009 11:29'!
fallbackFont
	"Answers the fallbackFont for the receiver. The fallback font must be some derivative of the receiver since it will not be asked to install itself properly on the target BitBlt so rendering a completely different font here is simply not possible. The default implementation uses a synthetic font that maps all characters to question marks."
	^ fallbackFont
		ifNil: [fallbackFont := FixedFaceFont new errorFont baseFont: self]! !

!TTCFont methodsFor: 'accessing' stamp: 'dgd 5/19/2007 16:20'!
fallbackFont: aFontSetOrNil

	aFontSetOrNil == self
		ifTrue:[^ self error: 'Invalid fallback font'].

	fallbackFont := aFontSetOrNil.
! !

!TTCFont methodsFor: 'private' stamp: 'topa 3/9/2015 23:27'!
fallbackForm
	"Compute the glyph form for the fallback glyph"
	^ttcDescription renderFallbackGlyphOfHeight: self height fgColor: foregroundColor bgColor: Color transparent depth: self depth! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 12/10/2002 17:08'!
familyName

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:39'!
familySizeFace

	^ Array
		with: self familyName
		with: self height
		with: 0.
! !

!TTCFont methodsFor: 'initialize' stamp: 'ar 11/14/2006 15:17'!
flushCache
	"Flush the cache of this font"
	cache := foregroundColor := colorToCacheMap := nil.! !

!TTCFont methodsFor: 'private' stamp: 'ar 8/29/2009 19:04'!
flushCachedValues
	"Flush all values computed from ttcDescription and cached for speed"
	"TTCFont allInstancesDo:[:font| font flushCachedValues]"
	height := ascent := descent := nil.! !

!TTCFont methodsFor: 'accessing' stamp: 'dgd 8/17/2004 22:10'!
fontNameWithPointSize
	^ self name withoutTrailingDigits , ' ' , self pointSize printString! !

!TTCFont methodsFor: 'public' stamp: 'yo 11/16/2002 00:59'!
foregroundColor

	^ foregroundColor.
! !

!TTCFont methodsFor: 'initialize' stamp: 'ar 6/2/2007 21:14'!
foregroundColor: fgColor
	"Install the given foreground color"
	foregroundColor = fgColor ifFalse:[
		foregroundColor := fgColor.
		colorToCacheMap ifNil:[colorToCacheMap := Dictionary new].
		cache := colorToCacheMap at: fgColor ifAbsentPut:[WeakArray new: self maxAscii+1].
		ShutdownList ifNotNil:[ShutdownList add: self].
	].
! !

!TTCFont methodsFor: 'private' stamp: 'ar 7/30/2009 21:47'!
formOf: char

	| code form |
	char charCode > self maxAscii
		ifTrue: [^ self fallbackFont formOf: char].

	cache ifNil:[self foregroundColor: Color black]. "make sure we have a cache"

	code := char charCode.
	form := cache at: (code + 1).
	form class == Association ifTrue:[^self computeForm: code]. "in midst of loading"
	form ifNil:[
		form := self computeForm: code.
		form ifNil:[^nil].
		cache at: code+1 put: form.
		GlyphCacheData at: (GlyphCacheIndex := GlyphCacheIndex \\ GlyphCacheSize + 1) put: form.
	].
	^form
! !

!TTCFont methodsFor: 'private' stamp: 'tpr 9/3/2013 13:58'!
glyphInfoOf: aCharacter into: glyphInfoArray
	"return the glyph info for aCharacter; if I don't have such a character, try my fallback font"

	| form |
	(self hasGlyphOf: aCharacter) ifFalse: [
		^ self fallbackFont glyphInfoOf: aCharacter into: glyphInfoArray.
	].
	form := self formOf: aCharacter.
	glyphInfoArray at: 1 put: form;
		at: 2 put: 0;
		at: 3 put: form width;
		at: 4 put: ascent "(self ascentOf: aCharacter)";
		at: 5 put: self.
	^ glyphInfoArray.
! !

!TTCFont methodsFor: 'private' stamp: 'topa 6/1/2016 11:24'!
hasGlyphOf: aCharacter
	"Answer whether this font includes a glyph for the given character (or a fallback)"
	self flag: #topa. "formOf: never actually returns nil for ttcfonts..."
	^ aCharacter charCode <= self maxAscii
		and:[(self formOf: aCharacter) notNil]! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/29/2009 18:17'!
height
	"Answer my height in pixels. This will answer a Float."
	^height ifNil:[height := self pixelSize + self lineGap]! !

!TTCFont methodsFor: 'html' stamp: 'dgd 9/2/2005 11:19'!
htmlSize
"private - answer the size in html form, assumes 12 as default size"
	^ (self pointSize - 12 / 3) rounded! !

!TTCFont methodsFor: 'private' stamp: 'topa 3/19/2015 02:43'!
indexOfSubfamilyName: aName
	
	^ self class indexOfSubfamilyName: aName! !

!TTCFont methodsFor: 'friend' stamp: 'yo 11/16/2002 01:00'!
initialize

	foregroundColor := Color black.
! !

!TTCFont methodsFor: 'initialize' stamp: 'tween 4/23/2006 21:45'!
initialize: aFont

	self initialize.
	self ttcDescription: aFont ttcDescription.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 1/6/2005 21:59'!
installOn: aDisplayContext

	^aDisplayContext installTTCFont: self.
! !

!TTCFont methodsFor: 'friend' stamp: 'xinwang 10/23/2010 20:49'!
installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor
	self foregroundColor: fgColor. "install color"
	(self fallbackFont notNil and: [ self fallbackFont class = TTCFontSet ]) ifTrue: [
		self fallbackFont fontArray do: [:font |
			font ifNotNil: [ font foregroundColor: fgColor ]
		]
	].  "install color for fallbackFont"
	aDisplayContext installTTCFont: self foregroundColor: foregroundColor backgroundColor: bgColor
! !

!TTCFont methodsFor: 'testing' stamp: 'yo 5/6/2004 19:25'!
isRegular
	"Answer true if I am a Regular/Roman font (i.e. not bold, etc.)"
	^ (self indexOfSubfamilyName: (self subfamilyName)) = 0.
! !

!TTCFont methodsFor: 'testing' stamp: 'nk 6/25/2003 12:55'!
isTTCFont
	^true! !

!TTCFont methodsFor: 'accessing' stamp: 'topa 4/3/2015 19:46'!
lineGap
	"Answer the line gap from the ttf description"
	^self pixelSize * ttcDescription typographicLineGap // ttcDescription typographicFontHeight! !

!TTCFont methodsFor: 'initialize' stamp: 'tween 4/23/2006 22:54'!
lineGlyph: ignore
	^self! !

!TTCFont methodsFor: 'accessing' stamp: 'dgd 12/11/2003 12:47'!
lineGrid
	"Answer the relative space between lines"
	^ self ascent + self descent! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 6/2/2007 21:12'!
maxAscii
	"Answer the max. code point in this font. The name of this method is historical."
	^maxAscii ifNil:[ttcDescription size].! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 6/2/2007 21:12'!
minAscii
	"Answer the min. code point in this font. The name of this method is historical."
	^ 0.
! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:00'!
name

	^ ttcDescription name.
! !

!TTCFont methodsFor: 'objects from disk' stamp: 'yo 3/17/2005 11:21'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a
reference to a known FontSet in the other system instead."

	"a path to me"
	dp := DiskProxy global: #TTCFont selector: #familyName:pointSize:emphasis:
			args: {self familyName. self pointSize. self emphasis}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTCFont methodsFor: 'html' stamp: 'dgd 9/2/2005 12:38'!
openHtmlOn: aStream 
	"put on the given stream the tag to open the html  
	representation of the receiver"
	| size |
	size := self htmlSize.
	size isZero
		ifFalse: [aStream nextPutAll: '<font size="' , size asString , '">']! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/29/2009 19:15'!
pixelScale
	"Answer the scale factor to convert from the truetype's units into the receiver's pixel size. The scale for a font is computed by *excluding* the linegap (leading) in the font. Consequently, the height of the font (including linegap/leading) will be larger than the pixel size."
	^self pixelSize asFloat / 
		(ttcDescription typographicAscender - 
			ttcDescription typographicLineGap - 
			ttcDescription typographicDescender)! !

!TTCFont methodsFor: 'accessing' stamp: 'ar 8/29/2009 18:26'!
pixelSize
	"Make sure that we don't return a Fraction"
	^(TextStyle pointsToPixels: pointSize) truncated! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 4/2/2004 11:27'!
pixelSize: aNumber
	"Make sure that we don't return a Fraction"
	self pointSize: (TextStyle pixelsToPoints: aNumber) rounded.
! !

!TTCFont methodsFor: 'as yet unclassified' stamp: 'topa 4/3/2015 20:04'!
pixelsPerInchChanged
	"The definition of TextStyle class>>pixelsPerInch has changed. Do whatever is necessary."
	self recreateCache! !

!TTCFont methodsFor: 'accessing' stamp: 'yo 6/23/2003 18:39'!
pointSize

	^ pointSize.
! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 7/18/2004 15:32'!
pointSize: aNumber

	self privatePointSize: aNumber.
	derivatives ifNotNil: [ derivatives do: [ :f | f ifNotNil: [ f privatePointSize: aNumber ]]].
! !

!TTCFont methodsFor: 'printing' stamp: 'yo 5/6/2004 19:25'!
printOn: aStream
	aStream nextPutAll: 'TTCFont(';
		nextPutAll: self familyName; space;
		print: self pointSize; space;
		nextPutAll: self subfamilyName;
		nextPut: $)! !

!TTCFont methodsFor: 'accessing' stamp: 'nk 7/18/2004 15:31'!
privatePointSize: aNumber 
	pointSize = aNumber
		ifFalse: [pointSize := aNumber.
			self flushCache]! !

!TTCFont methodsFor: 'friend' stamp: 'topa 3/10/2015 20:53'!
recreateCache

	cache := WeakArray new: 256.! !

!TTCFont methodsFor: 'caching' stamp: 'nk 3/15/2004 18:52'!
releaseCachedState
	self flushCache.! !

!TTCFont methodsFor: 'caching' stamp: 'yo 6/30/2004 14:49'!
reset
! !

!TTCFont methodsFor: 'private' stamp: 'nk 4/1/2004 09:15'!
scale

	^ self pixelSize / ttcDescription unitsPerEm
! !

!TTCFont methodsFor: 'emphasis' stamp: 'dgd 5/19/2007 16:19'!
setupDefaultFallbackFont

	| fonts f |
	fonts := TextStyle default fontArray.
	f := fonts first.
	1 to: fonts size do: [:i |
		self height > (fonts at: i) height ifTrue: [f := fonts at: i].
	].
	(f == self)
		ifFalse:[ self fallbackFont: f ].
	self reset.
! !

!TTCFont methodsFor: 'friend' stamp: 'yo 8/3/2005 07:44'!
setupDefaultFallbackFontTo: aTextStyleOrNil
"
	TTCFont allInstances do: [:i | i setupDefaultFallbackFontTo: (TextStyle named: 'MultiMSMincho')].
"

	| fonts f |
	aTextStyleOrNil ifNil: [
		self fallbackFont: nil.
		^ self.
	].
	fonts := aTextStyleOrNil fontArray.
	(aTextStyleOrNil defaultFont familyName endsWith: self familyName) ifTrue: [fallbackFont := nil. ^ self].

	f := fonts first.
	1 to: fonts size do: [:i |
		self height >= (fonts at: i) height ifTrue: [f := fonts at: i].
	].
	self fallbackFont: f.
	self reset.

! !

!TTCFont methodsFor: 'public' stamp: 'yo 12/10/2002 16:36'!
size

	^ ttcDescription size.
! !

!TTCFont methodsFor: 'private' stamp: 'yo 5/6/2004 19:23'!
subfamilyName

	^ ttcDescription subfamilyName.
! !

!TTCFont methodsFor: 'accessing' stamp: 'topa 3/10/2015 20:54'!
textStyle

	^ TextStyle actualTextStyles
		detect: [:aStyle | (aStyle fontArray collect: [:s | s name]) includes: self name]
		ifNone: [nil]! !

!TTCFont methodsFor: 'friend' stamp: 'yo 11/16/2002 01:01'!
ttcDescription

	^ ttcDescription.
! !

!TTCFont methodsFor: 'friend' stamp: 'ar 11/14/2006 15:19'!
ttcDescription: aTTCDescription

	ttcDescription := aTTCDescription.
	self flushCache.
! !

!TTCFont methodsFor: 'copying' stamp: 'yo 12/10/2002 16:36'!
veryDeepCopyWith: deepCopier

	self flushCache.
	^ self.
! !

!TTCFont methodsFor: 'public' stamp: 'ar 7/30/2009 21:48'!
widthOf: aCharacter
	"This method cannot use #formOf: because formOf: discriminates the color and causes unnecessary bitmap creation."
	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [^ fallbackFont widthOf: aCharacter].
		^ 1
	].
	^(self formOf: aCharacter) width! !

AbstractFont subclass: #TTCFontSet
	instanceVariableNames: 'name fontArray foregroundColor derivatives'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TrueType-Fonts'!

!TTCFontSet class methodsFor: 'private' stamp: 'yo 11/30/2002 23:47'!
discardDefault
"
	self discardDefault
"
	| ttc |
	ttc := TTCFontDescription default.
	ttc ifNotNil: [
		TextConstants removeKey: ttc name asSymbol ifAbsent: [].
	].! !

!TTCFontSet class methodsFor: 'instance creation' stamp: 'topa 3/18/2015 19:59'!
familyName: n pointSize: s

	"(self familyName: 'MSGothic' pointSize: 14) pointSize"
	| t ret index |
	t := self registry at: n asSymbol ifAbsent: [#()].
	t isEmpty ifTrue: [
		t := (TextConstants at: #DefaultTextStyle) fontArray.
		ret := t first.
		ret pointSize >= s ifTrue: [^ ret].
		index := 2.
		[index <= t size and: [(t at: index) pointSize <= s]] whileTrue: [
			ret := t at: index.
			index := index + 1.
		].
		^ ret.
	].
	^ (TextStyle named: n) addNewFontSize: s.! !

!TTCFontSet class methodsFor: 'instance creation' stamp: 'topa 3/18/2015 20:01'!
familyName: n pointSize: s emphasized: code


	| t ret index |
	t := self registry at: n asSymbol ifAbsent: [#()].
	t isEmpty ifTrue: [
		t := (TextConstants at: #DefaultTextStyle) fontArray.
		ret := t first.
		ret pointSize >= s ifTrue: [^ ret].
		index := 2.
		[index <= t size and: [(t at: index) pointSize <= s]] whileTrue: [
			ret := t at: index.
			index := index + 1.
		].
		^ ret emphasis: code
	].
	^ ((TextStyle named: n) addNewFontSize: s) emphasis: code.! !

!TTCFontSet class methodsFor: 'private' stamp: 'topa 3/10/2015 21:15'!
getExistings: fontArray

	| result em |
	result := OrderedCollection new.
	result add: fontArray.
	1 to: 3 do: [:i |
		em := (fontArray collect: [:f | f emphasized: i]).
		(em at: 1) ~= (fontArray at: 1) ifTrue: [
			result add: em.
		].
	].
	^ result asArray.
! !

!TTCFontSet class methodsFor: 'class initialization' stamp: 'topa 3/19/2015 11:50'!
initialize
"
	self initialize
"

	| tt |
	tt := TTCFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: {tt}].
! !

!TTCFontSet class methodsFor: 'private' stamp: 'yo 11/16/2002 01:18'!
newFontArray: anArray
 
	^super new initializeWithFontArray: anArray
! !

!TTCFontSet class methodsFor: 'file out/in' stamp: 'yo 12/2/2004 12:58'!
newTextStyleFromTTFile: fileName
"
	TTCFontReader encodingTag: JapaneseEnvironment leadingChar.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\msmincho.TTC'

	TTCFontReader encodingTag: 0.
	self newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\symbol.ttf'
"

	| description |
	description := TTCFontDescription addFromTTFile: fileName.
	^ self newTextStyleFromTT: description.
! !

!TTCFontSet class methodsFor: 'file out/in' stamp: 'topa 3/18/2015 20:04'!
newTextStyleFromTTFile: fileName encodingTag: encodingTag ranges: ranges

	| description |
	description := TTFontDescription addSetFromTTFile: fileName encodingTag: encodingTag ranges: ranges.
	self newTextStyleFromTT: description.
	^ description.
! !

!TTCFontSet class methodsFor: 'file out/in' stamp: 'topa 3/19/2015 02:55'!
newTextStyleFromTT: descriptionArray

	| array f arrayOfArray |

	arrayOfArray := self pointSizes collect: [:pt |
		descriptionArray collect: [:ttc |
			ttc ifNil: [nil] ifNotNil: [
				f := (ttc size > 256)
					ifTrue: [MultiTTCFont new]
					ifFalse: [TTCFont new].
				f ttcDescription: ttc.
				f pointSize: pt.
			].
		].
	].

	array := arrayOfArray collect: [:fonts |
		self newFontArray: fonts.
	].

	^TTCFont reorganizeForNewFontArray: array name: array first familyName asSymbol.
"
	styleName := (array at: 1) familyName asSymbol.
	textStyle := TextStyle fontArray: array.
	TextConstants at: styleName put: textStyle.

	self register: array at: styleName.

	^ TextConstants at: styleName.
"! !

!TTCFontSet class methodsFor: 'accessing' stamp: 'yo 11/30/2002 22:21'!
pointSizes

	^ TTCFont pointSizes.
! !

!TTCFontSet class methodsFor: 'registry' stamp: 'yo 12/2/2004 12:50'!
register: anObject at: symbolName

	self registry at: symbolName put: anObject.
! !

!TTCFontSet class methodsFor: 'registry' stamp: 'yo 12/7/2007 17:47'!
registry

	^ TTCFont registry.
! !

!TTCFontSet class methodsFor: 'registry' stamp: 'yo 12/2/2004 12:49'!
removeStyleName: aString

	| style symName |
	symName := aString asSymbol.
	style := TextConstants removeKey: symName ifAbsent: [].
	style ifNotNil: [self unregister: symName].
	TTCFontDescription removeDescriptionNamed: aString asString.
! !

!TTCFontSet class methodsFor: 'private' stamp: 'yo 12/1/2002 00:03'!
setDefault
"
	self setDefault
"
	| tt |
	tt := TTCFontDescription default.
	tt ifNil: [TTCFontDescription setDefault].
	tt := TTCFontDescription default.
	tt ifNotNil: [self newTextStyleFromTT: tt].
! !

!TTCFontSet class methodsFor: 'registry' stamp: 'yo 12/2/2004 12:49'!
unregister: symbolName

	self registry removeKey: symbolName ifAbsent: [].
! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'topa 3/19/2015 02:46'!
addLined: aTTCFont
	| l |
	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 4.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 16.
	self derivativeFont: l at: l emphasis.

	l := LinedTTCFont fromTTCFont: aTTCFont emphasis: 20.
	self derivativeFont: l at: l emphasis.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:15'!
ascent

	^ (fontArray at: 1) ascent.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/8/2004 21:18'!
ascentOf: aCharacter

	^ fontArray first ascentOf: aCharacter.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
baseKern

	^ 0.
! !

!TTCFontSet methodsFor: 'private' stamp: 'topa 6/1/2016 11:47'!
basicHasGlyphOf: aCharacter
	"see TTCFont>>hasGlyphOf:"
	^ fontArray first basicHasGlyphOf: aCharacter! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 12/10/2002 18:28'!
depth

	^ (fontArray at: 1) depth.
! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'topa 3/18/2015 20:12'!
derivativeFonts

	derivatives ifNil: [^ #()].
	^derivatives copyWithout: nil! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'topa 3/18/2015 20:12'!
derivativeFont: aTTCFont

	| index |
	index := TTCFont indexOfSubfamilyName: (aTTCFont subfamilyName).
	index < 1 ifTrue: [
		^ self "inform: 'unknown sub family name.  This font will be skipped'".
	].

	self derivativeFont: aTTCFont at: index.

	"self addLined: aTTCFont."
! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'topa 3/18/2015 20:13'!
derivativeFont: aTTCFontSet at: index

	| newDeriv |
	aTTCFontSet ifNil: [derivatives := nil. ^ self].
	derivatives ifNil: [derivatives := Array new: 32].
	derivatives size < 32 ifTrue: [
		newDeriv := Array new: 32.
		newDeriv replaceFrom: 1 to: derivatives size with: derivatives.
		derivatives := newDeriv.
	].
	derivatives at: index put: aTTCFontSet
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
descent

	^ (fontArray at: 1) descent.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
descentKern

	^ 0.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/8/2004 21:18'!
descentOf: aChar

	^ fontArray first descentOf: aChar
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'nice 1/10/2010 15:18'!
displayStringR2L: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta 

	| destPoint font form encoding char charCode glyphInfo |
	destPoint := aPoint.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do: [:charIndex |
		char := aString at: charIndex.
		encoding := char leadingChar + 1.
		charCode := char charCode.
		font := fontArray at: encoding.
		((charCode between: font minAscii and: font maxAscii) not) ifTrue: [
			charCode := font maxAscii].
		self glyphInfoOf: char into: glyphInfo.
		form := glyphInfo first.
			(glyphInfo size > 4 and: [glyphInfo fifth notNil and: [glyphInfo fifth ~= aBitBlt lastFont]]) ifTrue: [
				glyphInfo fifth installOn: aBitBlt.
			].
		aBitBlt
			sourceForm: form;
			destX: destPoint x - form width;
			destY: destPoint y;
			sourceOrigin: 0 @ 0;
			width: form width;
			height: form height;
			copyBits.
		destPoint := destPoint x - (form width + kernDelta) @ destPoint y.
	].
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'yo 1/7/2005 12:05'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta

	^ self displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.
! !

!TTCFontSet methodsFor: 'displaying' stamp: 'nice 1/10/2010 15:17'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

	| destPoint font form encoding glyphInfo char charCode destY |
	destPoint := aPoint.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do: [:charIndex |
		char := aString at: charIndex.
		encoding := char leadingChar + 1.
		charCode := char charCode.
		font := fontArray at: encoding.
		((charCode between: font minAscii and: font maxAscii) not) ifTrue: [
			charCode := font maxAscii].
		self glyphInfoOf: char into: glyphInfo.
		form := glyphInfo first.
		(glyphInfo fifth ~= aBitBlt lastFont) ifTrue: [
			glyphInfo fifth installOn: aBitBlt.
		].
		destY := baselineY - glyphInfo fourth. 
		aBitBlt
			sourceForm: form;
			destX: destPoint x;
			destY: destY;
			sourceOrigin: 0 @ 0;
			width: form width;
			height: form height;
			copyBits.
		destPoint := destPoint x + (form width + kernDelta) @ destPoint y.
	].
	^ destPoint.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'TN 3/14/2005 23:46'!
emphasis
	^ fontArray first emphasis! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'yo 12/7/2007 17:13'!
emphasis: code

	code > 3 ifTrue: [^ self].
	code = 0 ifTrue: [^ self].
	derivatives isNil ifTrue: [^ self].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 12/7/2007 17:13'!
emphasized: code

	code = 0 ifTrue: [^ self].
	derivatives ifNil: [^ self].
	(((code bitAnd: 20) ~= 0) and: [
		derivatives size < code or: [(derivatives at: code) isNil]]) ifTrue: [
		self addLined.
	].
	^ (derivatives at: code) ifNil: [self].
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 12/6/2007 13:43'!
familyName

	^ (fontArray at: 1) familyName.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
familySizeFace

	^ Array
		with: fontArray first name
		with: self height
		with: 0.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:16'!
fontArray

	^ fontArray
! !

!TTCFontSet methodsFor: 'private' stamp: 'tpr 9/3/2013 14:00'!
glyphInfoOf: aCharacter into: glyphInfoArray
"return glyph info for the character; it can get complicated."
	| index f code |
	"the leadingChar gives us an encoding. if that is outside the range of fonts in my fontArray, just return the default glyphInfo for a question-mark"
	index := aCharacter leadingChar + 1.
	fontArray size < index ifTrue: [^ self questionGlyphInfoInto: glyphInfoArray].
	(f := fontArray at: index) ifNil: [^ self questionGlyphInfoInto: glyphInfoArray].

	"if the actual character is not in the range supported by the chosen font, return that default question-mark info "
	code := aCharacter charCode.
	((code between: f minAscii and: f maxAscii) not) ifTrue: [
		^ self questionGlyphInfoInto: glyphInfoArray.
	].
	"finally, we have an ok font and character to return glyph info about - note that the font may have other ideas. "
	f glyphInfoOf: aCharacter into: glyphInfoArray.
	glyphInfoArray at: 5 put: self.
	^ glyphInfoArray.
! !

!TTCFontSet methodsFor: 'private' stamp: 'topa 6/1/2016 11:47'!
hasGlyphOf: aCharacter
	"see TTCFont>>hasGlyphOf:"
	^ fontArray first hasGlyphOf: aCharacter! !

!TTCFontSet methodsFor: 'accessing' stamp: 'topa 3/18/2015 20:15'!
hasGlyphWithFallbackOf: aCharacter

	| index f |
	index := aCharacter leadingChar +1.
	fontArray size < index ifTrue: [^ false].
	(f := fontArray at: index) ifNil: [^ false].

	^ f hasGlyphWithFallbackOf: aCharacter.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:27'!
height

	^fontArray first pixelSize.
! !

!TTCFontSet methodsFor: 'initialization' stamp: 'yo 12/10/2002 18:20'!
initializeWithFontArray: anArray

	fontArray := anArray.
	"name := anArray first name."
! !

!TTCFontSet methodsFor: 'displaying private' stamp: 'yo 1/6/2005 22:00'!
installOn: aDisplayContext

	^aDisplayContext installTTCFont: self.
! !

!TTCFontSet methodsFor: 'displaying private' stamp: 'yo 8/20/2003 22:51'!
installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor

	foregroundColor := fgColor.
	fontArray do: [:s | s ifNotNil: [s installOn: aDisplayContext foregroundColor: fgColor backgroundColor: bgColor]].
! !

!TTCFontSet methodsFor: 'testing' stamp: 'yo 2/12/2007 19:34'!
isFontSet

	^ true.
! !

!TTCFontSet methodsFor: 'testing' stamp: 'yo 12/8/2007 00:01'!
isRegular
	"Answer true if I am a Regular/Roman font (i.e. not bold, etc.)"
	^ (TTCFont indexOfSubfamilyName: (self subfamilyName)) = 0.
! !

!TTCFontSet methodsFor: 'testing' stamp: 'yo 12/29/2003 15:02'!
isTTCFont
	^true! !

!TTCFontSet methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:27'!
lineGrid

	^ fontArray first lineGrid.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:17'!
maxAsciiFor: encoding

	| f |
	f := (fontArray at: encoding+1).
	f ifNotNil: [^ f maxAscii].
	^ 0.
! !

!TTCFontSet methodsFor: 'objects from disk' stamp: 'tk 12/4/2004 14:51'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a
reference to a known FontSet in the other system instead."

	"a path to me"
	dp := DiskProxy global: #TTCFontSet selector: #familyName:pointSize:
			args: {self familyName. self pointSize}.
	refStrm replace: self with: dp.
	^ dp.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'topa 4/26/2015 22:58'!
pointSize

	^ fontArray first pointSize
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 11/16/2002 01:27'!
pointSizes

	^ self class pointSizes.
! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'yo 12/7/2007 17:12'!
pointSize: aNumber

	self privatePointSize: aNumber.
	derivatives ifNotNil: [ derivatives do: [ :f | f ifNotNil: [ f privatePointSize: aNumber ]]].
! !

!TTCFontSet methodsFor: 'derivatives' stamp: 'yo 12/7/2007 17:20'!
privatePointSize: aNumber
	fontArray do: [:f |
		f privatePointSize: aNumber
	].
! !

!TTCFontSet methodsFor: 'private' stamp: 'topa 3/18/2015 20:17'!
questionGlyphInfoInto: glyphInfoArray
"return glyph info for the question mark character in the first font of the fontArray -sort of a default set of info"
	| f form |
	f := fontArray at: 1.
	form := f formOf: $?.
	glyphInfoArray at: 1 put: form;
		at: 2 put: 0;
		at: 3 put: form width;
		at: 4 put: (self ascentOf: $?);
		at: 5 put: self.
	^ glyphInfoArray.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 12/7/2007 18:09'!
subfamilyName

	^ fontArray first ttcDescription subfamilyName.
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'yo 8/5/2003 15:31'!
textStyle

	^ TextStyle actualTextStyles
		detect: [:aStyle | (aStyle fontArray collect: [:s | s name]) includes: self name]
		ifNone: [].
! !

!TTCFontSet methodsFor: 'accessing' stamp: 'TN 3/14/2005 23:53'!
ttcDescription
	^ fontArray first ttcDescription! !

!TTCFontSet methodsFor: 'measuring' stamp: 'topa 3/10/2015 21:15'!
widthOf: aCharacter

	| encoding font |
	encoding := aCharacter leadingChar.
	encoding >= fontArray size ifFalse: [
		font := (fontArray at: encoding + 1).
		font ifNotNil: [^ font widthOf: aCharacter].
	].
	^ (fontArray at: 1) widthOf: aCharacter.
! !

!TextStyle methodsFor: '*TrueType' stamp: 'yo 5/7/2004 11:18'!
addLinedIfTT

	(fontArray first isKindOf: TTCFont) ifFalse: [^ self].

	fontArray do: [:f |
		f addLined.
	].
! !
TTFileDescription initialize!
TTFontDescription initialize!
TTFontReader initialize!
TTCFont initialize!
TTCFontSet initialize!
